"""
Module contains helper classes for popup widgets.

Popup widgets interact with server.

Server api:

  test(value)
      return element based on value criteria or empty string if no element meets criteria
      returned element is in format: (element id, string representation of element)

  dialog(value)
      return html form to choose element
      returned element is in format: (element id, string representation of element)
"""

import wx
from wx import ComboCtrl
from wx.lib import masked

from pytigon_lib.schtools import schjson
from pytigon_lib.schtools.tools import bencode, bdecode


class DataPopup(wx.ComboPopup):
    def __init__(self, size, combo, href):
        self.href = href
        self.combo = combo
        self.html = None
        self.size = size
        wx.ComboPopup.__init__(self)

    def Create(self, parent):
        self.html = self.combo.on_create(parent)
        parent.Bind(wx.EVT_KEY_DOWN, self.on_key_down)
        box = wx.BoxSizer(wx.VERTICAL)
        box.Add(self.html, 1, wx.ALL | wx.GROW, 1)
        parent.SetSizer(box)
        parent.SetAutoLayout(True)
        parent.Fit()
        return True

    def GetControl(self):
        return self.html

    def GetStringValue(self):
        return self.combo.start_value

    def on_key_down(self, event):
        if event.KeyCode == wx.WXK_ESCAPE:
            self.Dismiss()
        else:
            event.Skip()

    def set_new_href(self, href):
        self.href = href
        if self.html:
            self.html.set_new_href(href)

    def OnPopup(self):
        self.combo.on_popoup()
        return wx.ComboPopup.OnPopup(self)

    def GetAdjustedSize(self, minWidth, prefHeight, maxHeight):
        width = self.size[0]
        height = self.size[1]
        return wx.ComboPopup.GetAdjustedSize(self, width, height, maxHeight)


class DataPopupControl(ComboCtrl):
    def __init__(self, *args, **kwds):
        """Constructor"""
        if "style" in kwds:
            kwds["style"] |= wx.TE_PROCESS_ENTER
        else:
            kwds["style"] = wx.TE_PROCESS_ENTER

        if "dialog_with_value" in kwds:
            self.dialog_with_value = kwds["dialog_with_value"]
            del kwds["dialog_with_value"]
        else:
            self.dialog_with_value = True
        ComboCtrl.__init__(self, *args, **kwds)

        self.win = None
        self.html = None
        self.rec_value = []
        self.event_object = None
        self.page = None
        self.popup = None
        self.start_value = ""

        if self.defaultvalue:
            self.clear_str = self.defaultvalue
        else:
            self.clear_str = ""

        href = self.href.split(";")
        if len(href) > 1:
            self.href = href[0]
            self.href2 = href[1]
        else:
            self.href2 = None

        self.http = wx.GetApp().get_http(self)
        response = self.http.get(self, str(self.href) + "size/")
        self.size = schjson.loads(response.str())

        self.simpleDialog = True
        if self.GetTextCtrl():
            self.GetTextCtrl().SetForegroundColour(wx.Colour(0, 0, 0))

        self.UseAltPopupWindow(enable=True)

        popoup = self._create_popoup()
        self.SetPopupControl(popoup)

    def _create_popoup(self):
        if not self.popup:
            self.popup = DataPopup(size=self.size, combo=self, href=self.href)
        return self.popup

    def to_masked(self, **kwds):
        self.win = ComboCtrl.GetTextCtrl(self)
        self.win.__class__ = masked.TextCtrl
        self.win._PostInit(setupEventHandling=True, name="maskedTextCtrl", value="")
        self.win.SetCtrlParameters(**kwds)

    def GetTextCtrl(self):
        if self.win:
            return self.win
        return ComboCtrl.GetTextCtrl(self)

    def KillFocus(self):
        value = ComboCtrl.GetValue(self)
        if self.readonly:
            self.focus_out(value)

    def any_parent_command(self, command, *args, **kwds):
        parent = self
        while parent != None:
            if hasattr(parent, command):
                return getattr(parent, command)(*args, **kwds)
            parent = parent.GetParent()
        return None

    # def on_setfocus(self, event):
    #    value = ComboCtrl.GetValue(self)
    #    self.focus_in(value)
    #    event.Skip()

    def alternate_button_click(self):
        if self.event_object:
            if hasattr(self.event_object, "on_before_button_click"):
                self.event_object.on_before_button_click()

        self.run_ext_dialog()

        if self.event_object:
            if hasattr(self.event_object, "OnButtonClick"):
                self.event_object.OnButtonClick()

    def run_ext_dialog(self):
        """Run extended version of form to choose element"""
        self.GetTextCtrl().SetFocus()

        parm = dict()
        parm["value"] = self.get_parm("value")

        if self.href2:
            href = self.href2
        else:
            href = self.href

        _href = href + "dialog/|value" if self.dialog_with_value else href + "dialog/"
        self.page = self.GetParent().new_child_page(str(_href), "Select", parm)
        self.page.body.old_any_parent_command = self.page.body.any_parent_command
        self.page.body.any_parent_command = self.any_parent_command
        self.page.body.parent_combo = self

    def get_last_control_with_focus(self):
        return self

    def focus_in(self, value):
        pass

    def focus_out(self, value):
        if str(value) != self.start_value:
            if not str(value) == "":
                self.http = wx.GetApp().get_http(self)
                x = bencode(value)
                response = self.http.post(self, str(self.href) + "test/", {"value": x})
                tab = schjson.loads(self.response.str())
                ret = tab[0]

                if ret != 1:
                    if ret == 2:
                        self.OnButtonClick()
                    else:
                        self.clear_rec()
                else:
                    self.set_rec(tab[1], tab[2], False)

    def has_parm(self, parm):
        return True if parm == "value" else False

    def get_parm(self, parm):
        """For param = 'value' return field value bencoded"""
        return bencode(ComboCtrl.GetValue(self)) if parm == "value" else None

    def set_rec(self, value, value_rec, dismiss=False):
        """Set field value

        Args:
            value: element id
            value_rec: element
        """
        # if 'value' in value_rec.attrs:
        #    value2 = value_rec.data
        # else:
        #    value2 = value
        value2 = value_rec.data

        if self.event_object:
            if hasattr(self.event_object, "set_rec"):
                value2 = self.event_object.set_rec(value, value_rec, dismiss)

        self.start_value = value2
        self.SetValue(value2)
        self.rec_value = value_rec

        if dismiss:
            self.Dismiss()

        parent = self.GetParent()
        if hasattr(parent, "on_popup_control_change_value"):
            parent.on_popup_control_change_value(self)

    def get_rec(self):
        """Get element selected in field"""
        return self.rec_value

    def clear_rec(self):
        """Clear field"""
        self.start_value = ""
        self.SetValue(self.clear_str)
        self.rec_value = []

    def on_create(self, parent):
        from pytigon_gui.guiframe.page import SchPage

        href = (
            self.href + "dialog/|value"
            if self.dialog_with_value
            else self.href + "dialog/"
        )
        self.html = SchPage(parent, href, self)
        self.html.body.parent_combo = self
        return self.html

    def on_popoup(self):
        if self.html:
            wx.BeginBusyCursor()
            self.html.body.Hide()

            def _after():
                self.html.refresh_html()
                self.html.SetFocus()
                self.html.on_size(None)
                self.html.body.init()

                def _after2():
                    self.html.body.refr(self.start_value)
                    self.html.body.Show()

                wx.CallAfter(_after2)
                wx.EndBusyCursor()

            wx.CallAfter(_after)

    def Dismiss(self):
        if self.page:
            self.page.body.old_any_parent_command("on_cancel", None)
            self.page = None
        else:
            super().Dismiss()
        self.SetFocus()

    def set_new_href(self, href):
        """Set new base address to server service

        Args:
            href - new address
        """
        self.href = href

        href3 = self.href.split(";")
        if len(href3) > 1:
            self.href = href3[0]
            self.href2 = href3[1]
        else:
            self.href2 = None

        if self.href2:
            href3 = self.href2
        else:
            href3 = self.href

        if self.dialog_with_value:
            href3 += "dialog/|value"
        else:
            href3 += "dialog/"

        if self.popup:
            self.popup.set_new_href(href3)



import os
import wx
import importlib

from pytigon_lib.schtools.main_paths import get_main_paths

try:
    from pyshortcuts import make_shortcut
except:
    make_shortcut = None
import sys
import pytigon

_ = wx.GetTranslation


SIZE_DEFAULT = -1
SIZE_SMALL = 0
SIZE_MEDIUM = 1
SIZE_LARGE = 2


def norm_colour2(c):
    x = int(c)
    if x > 255:
        return 255
    else:
        return x


def norm_colour(r, g, b, proc):
    if proc > 1:
        y = (255, 255, 255)
        dx = 2 - proc
    else:
        y = (0, 0, 0)
        dx = proc
    dy = 1 - dx
    ret = [int(r * dx + y[0] * dy), int(g * dx + y[1] * dy), int(b * dx + y[1] * dy)]
    for i in range(3):
        if ret[i] > 255:
            ret[i] = 255
    return ret


def colour_to_html(colour):
    return wx.Colour(colour.Red(), colour.Green(), colour.Blue()).GetAsString(
        wx.C2S_HTML_SYNTAX
    )


def get_colour(wx_id, proc=None):
    c1 = wx.SystemSettings.GetColour(wx_id)
    if not proc:
        return colour_to_html(c1)
    else:
        x = norm_colour(c1.Red(), c1.Green(), c1.Blue(), proc)
        c2 = wx.Colour(x[0], x[1], x[2])
        return colour_to_html(c2)


def standard_tab_colour():
    return (
        ("color_body_0_2", get_colour(wx.SYS_COLOUR_3DFACE, 0.2)),
        ("color_body_0_5", get_colour(wx.SYS_COLOUR_3DFACE, 0.5)),
        ("color_body_0_7", get_colour(wx.SYS_COLOUR_3DFACE, 0.7)),
        ("color_body_0_9", get_colour(wx.SYS_COLOUR_3DFACE, 0.9)),
        ("color_body", get_colour(wx.SYS_COLOUR_3DFACE)),
        ("color_body_1_1", get_colour(wx.SYS_COLOUR_3DFACE, 1.1)),
        ("color_body_1_3", get_colour(wx.SYS_COLOUR_3DFACE, 1.3)),
        ("color_body_1_5", get_colour(wx.SYS_COLOUR_3DFACE, 0)),
        ("color_body_1_8", get_colour(wx.SYS_COLOUR_3DFACE, 1.8)),
        ("color_higlight", get_colour(wx.SYS_COLOUR_3DHIGHLIGHT)),
        ("color_shadow", get_colour(wx.SYS_COLOUR_3DSHADOW)),
        ("color_background_0_5", get_colour(wx.SYS_COLOUR_3DFACE, 0.5)),
        ("color_background_0_8", get_colour(wx.SYS_COLOUR_3DFACE, 0.8)),
        ("color_background_0_9", get_colour(wx.SYS_COLOUR_3DFACE, 0.9)),
        ("color_background", get_colour(wx.SYS_COLOUR_3DFACE)),
        ("color_background_1_1", get_colour(wx.SYS_COLOUR_3DFACE, 1.1)),
        ("color_background_1_2", get_colour(wx.SYS_COLOUR_3DFACE, 1.2)),
        ("color_background_1_5", get_colour(wx.SYS_COLOUR_3DFACE, 1.5)),
        ("color_info", get_colour(wx.SYS_COLOUR_INFOBK, 1.5)),
    )


def create_desktop_shortcut(app_name, title=None, parameters=""):
    pytigon_init_path = os.path.abspath(pytigon.__file__)
    ico_path = pytigon_init_path.replace("__init__.py", "pytigon.ico")
    ptig_path = pytigon_init_path.replace("__init__.py", "ptig.py")

    if "python" in sys.executable and make_shortcut:
        make_shortcut(
            ptig_path + " " + app_name, name=title if title else app_name, icon=ico_path
        )


LAST_FOCUS_CTRL_IN_FORM = None


def find_focus_in_form():
    global LAST_FOCUS_CTRL_IN_FORM
    win_focus = wx.Window.FindFocus()
    win = win_focus
    while win:
        if win.__class__.__name__ == "SchForm":
            LAST_FOCUS_CTRL_IN_FORM = win_focus
            return win_focus
        win = win.GetParent()
    if LAST_FOCUS_CTRL_IN_FORM and (
        not hasattr(LAST_FOCUS_CTRL_IN_FORM, "parent")
        or not LAST_FOCUS_CTRL_IN_FORM.parent
        or (
            hasattr(LAST_FOCUS_CTRL_IN_FORM.parent, "closing")
            and LAST_FOCUS_CTRL_IN_FORM.parent.closing
        )
    ):
        LAST_FOCUS_CTRL_IN_FORM = None
        return None
    else:
        return LAST_FOCUS_CTRL_IN_FORM


def import_plugin(plugin_name, prj_name=None):
    cfg = get_main_paths()
    pytigon_cfg = [cfg["PYTIGON_PATH"], "appdata", "plugins"]
    data_path = cfg["DATA_PATH"]
    data_cfg = [data_path, "plugins"]
    prj_cfg = [cfg["PRJ_PATH"], prj_name, "applib"]
    prj_cfg_alt = [cfg["PRJ_PATH_ALT"], prj_name, "applib"]

    if prj_name:
        folders = [prj_cfg, prj_cfg_alt]
    else:
        folders = [pytigon_cfg, data_cfg]

    path = None
    for folder in folders:
        plugins_path = os.path.join(folder[0], *folder[1:])
        if prj_name:
            plugin_path = os.path.join(plugins_path, *plugin_name.split(".")[:-1])
        else:
            plugin_path = os.path.join(plugins_path, *plugin_name.split("."))
        if os.path.exists(plugin_path):
            path = plugins_path
            path2 = plugin_path
            break

    if not path:
        return None

    try:
        m = importlib.import_module(plugin_name, package=None)
        return m
    except:
        try:
            m = importlib.import_module(plugin_name, package=None)
            return m
        except:
            pass
    return None



"""
This module contains SchAppFrame class. When pytigon application starts, new top frame window (class SchAppFrame) is created.
During the initialisation process all defined for application plugins are started.
"""

import os
import sys
import platform
import subprocess
import base64
from pydispatch import dispatcher
from pathlib import Path

import wx

# import wx.html2
import datetime
from wx.lib.agw import aui

from django.conf import settings

from pytigon_lib.schfs.vfstools import get_temp_filename
from pytigon_lib.schdjangoext.tools import gettempdir
from pytigon_lib.schhttptools.httpclient import HttpResponse
from pytigon_lib import schfs

from pytigon_lib.schtools.tools import split2

# from pytigon_lib.schtasks.task import get_process_manager

from pytigon_gui.guilib.image import ArtProviderFromIcon
from pytigon_gui.guilib.events import *  # @UnusedWildImport
from pytigon_gui.guiframe.notebook import SchNotebook
from pytigon_gui.guiframe.notebookpage import SchNotebookPage
from pytigon_gui.guiframe.manager import SChAuiManager
from pytigon_gui.guilib.image import bitmap_from_href

from pytigon_gui.guiframe.baseframe import SchBaseFrame

_ = wx.GetTranslation

_RECORD_VIDEO = 0
_RECORD_VIDEO_STRUCT = None
_RECORD_VIDEO_OUT = None
_RECORD_VIDEO_MONITOR = None


def save_video_frame(win):
    global _RECORD_VIDEO_STRUCT, _RECORD_VIDEO_OUT, _RECORD_VIDEO_MONITOR
    if _RECORD_VIDEO_STRUCT:
        mss = _RECORD_VIDEO_STRUCT[0]
        cv2 = _RECORD_VIDEO_STRUCT[1]
        numpy = _RECORD_VIDEO_STRUCT[2]
        if not _RECORD_VIDEO_OUT:
            pos = win.GetScreenPosition()
            size = win.GetSize()

            # fourcc = cv2.VideoWriter_fourcc(*'MJPG')
            fourcc = cv2.VideoWriter_fourcc(*"XVID")
            _RECORD_VIDEO_OUT = cv2.VideoWriter(
                win._video, fourcc, 4, (size.width, size.height)
            )
            _RECORD_VIDEO_MONITOR = {
                "top": pos.y,
                "left": pos.x,
                "width": size.width,
                "height": size.height,
            }

        with mss() as sct:
            pos = win.GetScreenPosition()
            size = win.GetSize()
            frame = numpy.array(sct.grab(_RECORD_VIDEO_MONITOR))
            frame = numpy.flip(frame[:, :, :3], 2)
            frame = cv2.cvtColor(frame, cv2.COLOR_BGR2RGB)
            _RECORD_VIDEO_OUT.write(frame)


def finish_video():
    if _RECORD_VIDEO_OUT:
        _RECORD_VIDEO_OUT.release()
        _RECORD_VIDEO_STRUCT = None


class _SChMainPanel(wx.Window):
    def __init__(self, app_frame, *argi, **argv):
        argv["name"] = "schmainpanel"
        if "style" in argv:
            argv["style"] |= wx.WANTS_CHARS
        self.app_frame = app_frame
        wx.Window.__init__(self, *argi, **argv)
        self.SetBackgroundColour(wx.SystemSettings.GetColour(wx.SYS_COLOUR_BACKGROUND))

    def GetFrameManager(self):
        return self.GetParent().GetFrameManager()

    def GetMenuBar(self):
        return self.GetParent().GetMenuBar()

    def Freeze(self):
        pass

    def Thaw(self):
        pass


class SchAppFrame(SchBaseFrame):
    """
    This is main window of pytigon application
    """

    def __init__(
        self,
        gui_style="tree(toolbar,statusbar)",
        title="",
        pos=wx.DefaultPosition,
        size=wx.DefaultSize,
        style=wx.DEFAULT_FRAME_STYLE | wx.CLIP_CHILDREN | wx.WANTS_CHARS,
        video_name="",
    ):
        """Constructor

        Args:
            gui_style - there is string with some key words. If specyfied key word exist some functionality of
            SChAppFrame are turned on.

            List of key words:

                dialog - application with one form

                one_form - some as dialog

                tray - application have system tray icon with some actions

                statusbar - show the status bar

                tree - menu in tree form

                standard - standard for platform interface

                modern - replace standard toolbar with ribbon bar

                toolbar - show the tool bar

            title - the window title

            pos - the window position

            size - the window size

            style - see wx.Frame constructor
        """

        SchBaseFrame.__init__(
            self, None, gui_style, wx.ID_ANY, title, pos, size, style, "MainWindow"
        )
        # self.SetBackgroundColour(wx.SystemSettings.GetColour(wx.SYS_COLOUR_BACKGROUND))

        self._id = wx.ID_HIGHEST
        self._perspectives = []
        self._menu_bar_lp = 0
        self._toolbar_bar_lp = 1
        self._proc_mannager = None
        self._video = video_name
        self.gui_style = gui_style
        self.idle_objects = []
        self.gui_style = gui_style
        self.command = {}
        # self.command_enabled_always = []
        self.last_pane = None
        self.active_pane = None
        self.active_page = None
        self.active_child_ctrl = None

        self.toolbar_interface = None
        self.menubar_interface = None
        self.statusbar_interface = None

        self.destroy_fun_tab = []
        self.after_init = False

        self.sys_command = dict({"EXIT": self.on_exit, "ABOUT": self.on_about})

        if "dialog" in self.gui_style or "one_form" in self.gui_style:
            hide_on_single_page = True
        else:
            hide_on_single_page = False

        self._panel = _SChMainPanel(self, self)
        self._mgr = SChAuiManager()
        self._mgr.SetManagedWindow(self._panel)

        if not hasattr(self._mgr, "SetAGWFlags"):
            self._mgr.SetAGWFlags = self._mgr.SetFlags
            self._mgr.GetAGWFlags = self._mgr.GetFlags

        self._mgr.SetAGWFlags(self._mgr.GetAGWFlags() ^ aui.AUI_MGR_ALLOW_ACTIVE_PANE)

        self.desktop = self._create_notebook_ctrl(hide_on_single_page)
        self.get_dock_art().SetMetric(aui.AUI_DOCKART_PANE_BORDER_SIZE, 0)

        wx.ArtProvider.Push(ArtProviderFromIcon())

        icon = wx.Icon()
        b = wx.Bitmap(wx.Image(wx.GetApp().src_path + "/static/icons/schweb.png"))
        icon.CopyFromBitmap(b)
        self.SetIcon(icon)

        if "tray" in gui_style:
            parent = self

            class _TaskBarIcon(wx.adv.TaskBarIcon):
                def CreatePopupMenu(self):
                    nonlocal parent
                    parent.PopupMenu(parent.menu_tray)
                    return None

            self.tbIcon = _TaskBarIcon()
            self.tbIcon.SetIcon(icon, "Pytigon")
        else:
            self.tbIcon = None

        if "statusbar" in gui_style:
            self.statusbar = self._create_status_bar()
        else:
            self.statusbar = None

        wx.GetApp().SetTopWindow(self)

        self.setup_frame()

        if "tree" in gui_style:
            self._sizer = wx.BoxSizer(wx.HORIZONTAL)
        else:
            self._sizer = wx.BoxSizer(wx.VERTICAL)

        if "standard" in gui_style:
            if len(wx.GetApp().get_tab(self._toolbar_bar_lp)) > 1:
                from pytigon_gui.toolbar import standardtoolbar

                self.toolbar_interface = standardtoolbar.StandardToolbarBar(
                    self, gui_style
                )
                self._create_tool_bar()
                self.toolbar_interface.create()
            self._sizer.Add(self._panel, 1, wx.EXPAND)
            self._mgr.Update()
        elif "generic" in gui_style:
            if len(wx.GetApp().get_tab(self._toolbar_bar_lp)) > 1:
                from pytigon_gui.toolbar import generictoolbar

                self.toolbar_interface = generictoolbar.GenericToolbarBar(
                    self, gui_style
                )
                self._create_tool_bar()
                self.toolbar_interface.create()
            self._sizer.Add(self._panel, 1, wx.EXPAND)
            self._mgr.Update()
        elif "modern" in gui_style:
            from pytigon_gui.toolbar.moderntoolbar import ModernToolbarBar

            self.toolbar_interface = ModernToolbarBar(self, gui_style)
            self._create_tool_bar()
            # self.toolbar_interface.realize_bar()
            self.toolbar_interface.create()
            self._sizer.Add(self.toolbar_interface, 0, wx.EXPAND)
            self._sizer.Add(self._panel, 1, wx.EXPAND)
        elif "tree" in gui_style:
            from pytigon_gui.toolbar.treetoolbar import TreeToolbarBar

            self.toolbar_interface = TreeToolbarBar(self._panel, gui_style)
            self._create_tool_bar()
            self.toolbar_interface.create()
            self._mgr.AddPane(
                self.toolbar_interface,
                self._create_pane_info("menu", _("Menu"))
                .CaptionVisible(True)
                .MinimizeButton(True)
                .CloseButton(False)
                .Left()
                .BestSize((250, 40))
                .MinSize((250, 40))
                .Show(),
            )
            self._sizer.Add(self._panel, 1, wx.EXPAND)
        else:
            self._sizer.Add(self._panel, 1, wx.EXPAND)

        self.SetSizer(self._sizer)

        if wx.GetApp().get_tab(2) and len(wx.GetApp().get_tab(2)) > 1:
            self._menu_bar_lp = 2
        else:
            if wx.GetApp().menu_always:
                self._menu_bar_lp = 1

        if self._menu_bar_lp > 0:
            from pytigon_gui.toolbar import menubar

            self.menubar_interface = menubar.MenuToolbarBar(self, gui_style)
            self._create_menu_bar()

        s_dx = wx.SystemSettings.GetMetric(wx.SYS_SCREEN_X)
        s_dy = wx.SystemSettings.GetMetric(wx.SYS_SCREEN_Y)

        self._mgr.AddPane(
            self._create_notebook_ctrl(),
            self._create_pane_info("panel", _("Tools"))
            .CaptionVisible(True)
            .Left()
            .MinSize((400, int(s_dy / 2)))
            .BestSize((int(s_dx / 2) - 50, s_dy - 100))
            .Show(),
        )
        self._mgr.AddPane(
            self._create_notebook_ctrl(),
            self._create_pane_info("header", _("Header"))
            .CaptionVisible(False)
            .Top()
            .MinSize((int(s_dx), int(s_dy / 10)))
            .BestSize((int(s_dx), int(s_dy / 5)))
            .Show(),
        )
        self._mgr.AddPane(
            self._create_notebook_ctrl(),
            self._create_pane_info("footer", _("Footer"))
            .CaptionVisible(False)
            .Bottom()
            .MinSize((int(s_dx), int(s_dy / 10)))
            .BestSize((int(s_dx), int(s_dy / 5)))
            .Show(),
        )
        self._mgr.AddPane(
            self.desktop,
            self._create_pane_info("desktop", _("Desktop")).CenterPane().Show(),
        )
        perspective_notoolbar = self._mgr.SavePerspective()

        if "toolbar" in gui_style and "standard" in gui_style:
            i = 1
            while True:
                name = "tb" + str(i)
                tbpanel = self._mgr.GetPane(name)
                if tbpanel.IsOk():
                    tbpanel.Show()
                else:
                    break
                i += 1

        perspective_all = self._mgr.SavePerspective()
        all_panes = self._mgr.GetAllPanes()

        for ii in range(len(all_panes)):
            if not all_panes[ii].IsToolbar():
                if all_panes[ii].name != "menu":
                    all_panes[ii].Hide()

        self._mgr.GetPane("desktop").Show()

        perspective_default = self._mgr.SavePerspective()

        size = self.desktop.GetPageCount()
        for i in range(size):  # @UnusedVariable
            self.desktop.DeletePage(0)

        self._perspectives.append(perspective_default)
        self._perspectives.append(perspective_all)
        self._perspectives.append(perspective_notoolbar)

        self.init_acc_keys()
        self.init_plugins()
        self.SetAcceleratorTable(wx.AcceleratorTable(self.aTable))

        self._mgr.Update()

        self.t1 = wx.Timer(self)
        self.t1.Start(250)

        self.Bind(wx.EVT_TIMER, self.on_timer, self.t1)
        self.Bind(aui.EVT_AUI_PANE_ACTIVATED, self.on_pane_activated)
        self.Bind(
            wx.EVT_MENU_RANGE, self.on_show_elem, id=ID_SHOWHEADER, id2=ID_SHOWTOOLBAR2
        )

        if "tray" in gui_style:
            self.Bind(wx.EVT_CLOSE, self.on_taskbar_hide)
        else:
            self.Bind(wx.EVT_CLOSE, self.on_close)

        self._panel.Bind(wx.EVT_CHILD_FOCUS, self.on_child_focus)

        self.bind_command(self.on_open, id=wx.ID_OPEN)
        self.bind_command(self.on_exit, id=wx.ID_EXIT)

        # if self.toolbar_interface:
        # self.toolbar_interface.bind_ui(self.on_update_ui_command, wx.ID_ANY)

        # for pos in [wx.ID_EXIT, ID_WEB_NEW_WINDOW, wx.ID_OPEN]:
        #    self.command_enabled_always.append(pos)

        self.bind_command(self.on_next_tab, id=ID_NEXTTAB)
        self.bind_command(self.on_prev_tab, id=ID_PREVTAB)
        self.bind_command(self.on_next_page, id=ID_NEXTPAGE)
        self.bind_command(self.on_prev_page, id=ID_PREVPAGE)
        self.bind_command(self.on_close_tab, id=ID_CLOSETAB)
        self.bind_command(self.on_refresh_tab, id=ID_REFRESHTAB)

        self.bind_command(self.on_goto_desktop, id=ID_GOTODESKTOP)
        self.bind_command(self.on_goto_head, id=ID_GOTOHEADER)
        self.bind_command(self.on_goto_panel, id=ID_GOTOPANEL)
        self.bind_command(self.on_goto_footer, id=ID_GOTOFOOTER)

        self.bind_command(self.on_command, id=ID_WEB_NEW_WINDOW)

        self.bind_command(self.on_show_status_bar, id=ID_SHOWSTATUSBAR)

        if self.menubar_interface:
            self.SetMenuBar(self.menubar_interface)

        if self.tbIcon:
            # self.tbIcon.Bind(wx.adv.EVT_TASKBAR_LEFT_DCLICK, self.on_taskbar_toogle)
            # self.tbIcon.Bind(wx.adv.EVT_TASKBAR_RIGHT_UP, self.on_taskbar_show_menu)

            self.menu_tray = wx.Menu()
            self.menu_tray.Append(ID_TASKBAR_SHOW, _("Show It"))
            self.Bind(wx.EVT_MENU, self.on_taskbar_show, id=ID_TASKBAR_SHOW)
            self.menu_tray.Append(ID_TASKBAR_HIDE, _("Minimize It"))
            self.Bind(wx.EVT_MENU, self.on_taskbar_hide, id=ID_TASKBAR_HIDE)
            self.menu_tray.AppendSeparator()
            self.menu_tray.Append(ID_TASKBAR_CLOSE, _("Close It"))
            self.Bind(wx.EVT_MENU, self.on_close, id=ID_TASKBAR_CLOSE)

        self.SetExtraStyle(wx.WS_EX_PROCESS_UI_UPDATES)
        wx.UpdateUIEvent.SetUpdateInterval(50)
        wx.UpdateUIEvent.SetMode(wx.UPDATE_UI_PROCESS_SPECIFIED)
        wx.CallAfter(self.UpdateWindowUI)
        self.Layout()
        wx.CallAfter(self.Bind, wx.EVT_IDLE, self.on_idle)
        wx.CallAfter(self.Show)

    def setup_frame(self):
        self.SetMinSize(wx.Size(800, 600))

    def init_acc_keys(self):
        self.aTable = [
            (wx.ACCEL_ALT, wx.WXK_PAGEUP, ID_PREVTAB),
            (wx.ACCEL_ALT, wx.WXK_PAGEDOWN, ID_NEXTTAB),
            (wx.ACCEL_ALT, wx.WXK_LEFT, ID_GOTOPANEL),
            (wx.ACCEL_ALT, ord("t"), ID_GOTOPANEL),
            (wx.ACCEL_ALT, wx.WXK_RIGHT, ID_GOTODESKTOP),
            (wx.ACCEL_ALT, ord("d"), ID_GOTODESKTOP),
            (wx.ACCEL_ALT, wx.WXK_UP, ID_GOTOHEADER),
            (wx.ACCEL_CTRL, ord("w"), ID_CLOSETAB),
            (wx.ACCEL_CTRL, ord("n"), ID_WEB_NEW_WINDOW),
            (wx.ACCEL_CTRL, wx.WXK_TAB, ID_NEXTTAB),
            (wx.ACCEL_CTRL | wx.ACCEL_SHIFT, wx.WXK_TAB, ID_PREVTAB),
            (wx.ACCEL_CTRL, wx.WXK_F6, ID_NEXTTAB),
            (wx.ACCEL_CTRL | wx.ACCEL_SHIFT, wx.WXK_F6, ID_PREVTAB),
            (wx.ACCEL_ALT, ord("h"), ID_PREVTAB),
            (wx.ACCEL_ALT, ord("l"), ID_NEXTTAB),
            (wx.ACCEL_ALT, ord("k"), ID_PREVPAGE),
            (wx.ACCEL_ALT, ord("j"), ID_NEXTPAGE),
            (wx.ACCEL_ALT, wx.WXK_BACK, ID_WEB_BACK),
            (0, wx.WXK_F5, ID_REFRESHTAB),
        ]

    def on_idle(self, event):
        for obj in self.idle_objects:
            obj.on_idle()

        if not self.after_init:
            self.after_init = True
            app = wx.GetApp()
            if len(app.start_pages) > 0:

                def start_pages():
                    for page in app.start_pages:
                        url_page = page.split(";")
                        if len(url_page) == 2:
                            self._on_html(
                                _(url_page[0]) + "," + app.base_path + "/" + url_page[1]
                            )
                        elif len(url_page) == 1:
                            self._on_html("," + app.base_path + "/" + url_page[0])

                wx.CallAfter(start_pages)

        event.Skip()

    def on_pane_activated(self, event):
        active_pane = event.GetPane()

        if self.active_pane != active_pane:
            if self.active_pane and hasattr(self.active_pane, "set_active"):
                self.active_pane.set_active(False)
                if self.active_pane.GetCurrentPage():
                    self.active_pane.GetCurrentPage().Refresh()

            self.active_pane = active_pane

            if active_pane and hasattr(active_pane, "set_active"):
                active_pane.set_active(True)
                if active_pane.GetCurrentPage():
                    active_pane.GetCurrentPage().Refresh()

        event.Skip()

    def get_frame_manager(self):
        """returns frame manager - aui.framemanager.AuiManager derived object"""
        return self._mgr

    def get_dock_art(self):
        """return art provider related to the asociated frame manager"""
        return self._mgr.GetArtProvider()

    def SetActive(self, notebook, tab):
        if tab:
            tab.SetFocus()

    def _create_notebook_ctrl(self, hideSingleTab=True):
        style = (
            aui.AUI_NB_WINDOWLIST_BUTTON
            | aui.AUI_NB_CLOSE_ON_ALL_TABS
            | aui.AUI_NB_TAB_MOVE
            | aui.AUI_NB_TAB_EXTERNAL_MOVE
            | aui.AUI_NB_TAB_SPLIT
            | aui.AUI_NB_WINDOWLIST_BUTTON
        )
        n = SchNotebook(self._panel, wx.Point(0, 0), wx.Size(0, 0), style=style)
        n.SetAGWWindowStyleFlag(style)
        n.SetArtProvider(aui.VC71TabArt())
        return n

    def _create_pane_info(self, name, caption):
        return aui.AuiPaneInfo().Name(name).Caption(caption)

    def _on_page_event(self, direction):
        """scroll active form"""
        w = wx.Window.FindFocus()
        parent = w
        while parent:
            if parent.__class__.__name__ == "SchForm":
                if hasattr(parent, "on_page_event"):
                    parent.on_page_event(direction)
                    return
                else:
                    y = parent.GetViewStart()[1] * parent.GetScrollPixelsPerUnit()[1]
                    dy = (
                        parent.GetScrollPageSize(wx.VERTICAL)
                        * parent.GetScrollPixelsPerUnit()[1]
                    )
                    y = y + direction * dy
                    if y < 0:
                        y = 0
                    if parent.GetScrollPixelsPerUnit()[1] != 0:
                        parent.Scroll(-1, y / parent.GetScrollPixelsPerUnit()[1])
                    else:
                        parent.Scroll(-1, y)
                    return
            else:
                parent = parent.GetParent()

    def on_next_page(self, evt):
        """scroll active form one page down"""
        self._on_page_event(1)

    def on_prev_page(self, evt):
        """scroll active form one page up"""
        self._on_page_event(-1)

    def on_next_tab(self, evt):
        """select a next tab in desktop notebook"""
        desktop = wx.GetApp().GetTopWindow().desktop
        id = desktop.GetPageIndex(desktop.GetCurrentPage())
        idn = 0
        if id != None and id >= 0:
            idn = id + 1
        if idn >= desktop.GetPageCount():
            idn = 0
        desktop.SetSelection(idn)
        return

    def on_prev_tab(self, evt):
        """select a previous tab in desktop notebook"""
        desktop = wx.GetApp().GetTopWindow().desktop
        id = desktop.GetSelection()
        idn = 0
        if id >= 0:
            idn = id - 1
        if idn < 0:
            idn = desktop.GetPageCount() - 1
        desktop.SetSelection(idn)
        return

    def on_close_tab(self, evt):
        """close active tab in active notebook"""
        win = wx.Window.FindFocus()
        while win:
            if win.__class__.__name__ == "SchNotebook":
                id = win.GetSelection()
                if id >= 0:
                    panel = win.GetPage(id)
                    panel.close_child_page(True)
                return
            win = win.GetParent()

    def after_close(self, win):
        count = win.GetPageCount()
        if count < 1:
            apppanel = win.GetPanel()
            if apppanel:
                apppanel.Hide()
                self._mgr.Update()

    def on_refresh_tab(self, evt):
        """Refresh content of active form"""
        desktop = wx.GetApp().GetTopWindow().desktop
        idn = desktop.GetSelection()
        if idn >= 0:
            panel = desktop.GetPage(idn)
            panel.refresh_html()

    def get_active_ctrl(self):
        """Get active widget"""
        ctrl = None
        idn = self.desktop.GetSelection()
        if idn >= 0:
            panel = self.desktop.GetPage(idn)
            if panel:
                count = panel.get_page_count()
                if count > 0:
                    htmlwin = panel.get_page(count - 1)
                    if htmlwin:
                        ctrl = htmlwin.get_last_control_with_focus()
                return ctrl
        return None

    def get_active_panel(self):
        """get active form"""
        idn = self.desktop.GetSelection()
        if idn >= 0:
            panel = self.desktop.GetPage(idn)
            if panel:
                count = panel.GetPageCount()
                if count > 0:
                    htmlwin = panel.GetPage(count - 1)
                    return htmlwin
        return None

    def _open_page(
        self,
        response,
        title="",
        parameters=None,
        view_in=None,
    ):
        address = response.url
        # html: text/html
        # js: text/javascript
        # jsnon: application/json
        if "text/" in response.ret_content_type:
            return self.new_main_page(response, title, parameters, view_in)

        # html: text/python
        if "text/python" in response.ret_content_type:
            exec(http.str())
            return

        # pdf: "application/pdf"
        if "application/pdf" in response.ret_content_type:
            return self.show_pdf(response, title, parameters)

        # spdf: "application/spdf"
        if "application/spdf" in response.ret_content_type:
            if "?print=" in address:
                return self.print_spdf(response, title, parameters)
            else:
                return self.show_spdf(response, title, parameters)
            # return self.show_spdf(response, title, parameters)

        # ods: "application/vnd.oasis.opendocument.spreadsheet"
        # odp: "application/vnd.oasis.opendocument.text"
        # xlsx: "application/vnd.openxmlformats-officedocument.spreadsheetml.sheet"
        # docx: "application/vnd.openxmlformats-officedocument.wordprocessingml.document

        if "application/vnd" in response.ret_content_type:
            return self.show_document(response, title, parameters)

        if "application/" in response.ret_content_type:
            return self.download_data(response, title, parameters)

        if (
            "video/" in response.ret_content_type
            or "audio/" in response.ret_content_type
        ):
            return self.show_document(response, title, parameters)

        # "image/"
        if "image/" in response.ret_content_type:
            return self.show_image(response, title, parameters)

        # zip: application/zip
        if "application/zip" in response.ret_content_type:
            return self.download_data(response, title, parameters)

        if "application/json" in response.ret_content_type:
            return self.download_data(response, title, parameters, view_in)

        return self.download_data(response, title, parameters)

    def open_page(
        self,
        address,
        title="",
        parameters=None,
        view_in=None,
    ):
        http = wx.GetApp().get_http_for_adr(address)
        if parameters:
            response = http.post(self, address, parameters)
        else:
            response = http.get(self, address)

        return self._open_page(response, title, parameters, view_in)

    def new_main_page(
        self, address_or_parser, title="", parameters=None, view_in="desktop", http=None
    ):
        """Open a new page in main

        Args:
            address_or_parser: can be: address of http page (str type) or
            :class:'~pytigon_lib.schparser.html_parsers.ShtmlParser'
            title - new tab title
            parameters - parameters of http request
            view_in - options are: 'desktop', 'panel', 'header', 'footer'
        Returns:
            created form :class:'~pytigon_gui.guiframe.htmlsash.SchPage'
        """

        if type(address_or_parser) == str:
            address = address_or_parser
            if (
                not address.startswith("^")
                and not address.startswith("file://")
                and view_in == "desktop"
            ):
                return self.open_page(address_or_parser, title, parameters, view_in)

        elif type(address_or_parser) == HttpResponse:
            address = address_or_parser.url
        else:
            address = address_or_parser.address

        ext = address.split("?")[0].split(".")[-1]
        if (
            (not view_in or not "browser" in view_in)
            and ext
            and ext
            in (
                "pdf",
                "spdf",
                "ods",
                "odp",
                "odt",
                "docx",
                "xlsx",
                "pptx",
                "zip",
                "avi",
                "mpeg",
                "mp3",
                "mp4",
            )
        ):
            return self.open_page(address, title, parameters, view_in)

        if view_in:
            panel = view_in
        else:
            parm = split2(address, "?")
            if len(parm) == 2:
                panel = "desktop"
                parm2 = parm[1].split(",")
                parm3 = [pos.split("=") for pos in parm2]
                for pos in parm3:
                    if len(pos) == 2:
                        if pos[0] == "show_in":
                            panel = pos[1]
                            break

        if panel == "pscript":
            if not http:
                http = wx.GetApp().get_http(self)
            response = http.get(self, address)
            ptr = response.str()
            exec(ptr)
            return

        if not address.startswith("^"):
            if (not view_in or view_in.startswith("browser")) or (
                address.startswith("http")
                and not address.startswith(wx.GetApp().base_address)
            ):
                if view_in and "_" in view_in:
                    panel = view_in.split("_")[1]
                else:
                    panel = "desktop"
                ret = self.new_main_page(
                    "^standard/webview/widget_web.html", "Empty page", view_in=panel
                )
                if (
                    address.startswith("http://")
                    or address.startswith("https://")
                    or address.startswith("file://")
                ):

                    def _ret_fun():
                        if type(address_or_parser) == str:
                            ret.body.WEB.go(address)
                        else:
                            ret.body.WEB.load_str(address_or_parser.ptr())

                    wx.CallAfter(_ret_fun)
                else:

                    def _ret_fun():
                        if type(address_or_parser) == str:
                            ret.body.WEB.go(wx.GetApp().base_path + address)
                        else:
                            ret.body.WEB.load_str(address_or_parser.ptr())

                    wx.CallAfter(_ret_fun)
                return ret

        if len(title) < 32:
            title2 = title
        else:
            title2 = title[:30] + "..."

        if panel.startswith("toolbar"):
            name = panel[7:]
            if name[0:1] == "_":
                return self.toolbar_interface.create_html_win(
                    name[1:], address_or_parser, parameters
                )
            else:
                return self.toolbar_interface.create_html_win(
                    None, address_or_parser, parameters
                )
        if panel == "desktop2":
            n = self._mgr.GetPane("desktop").window
        else:
            n = self._mgr.GetPane(panel).window

        if not self._mgr.GetPane(panel).IsShown():
            refr = True
        else:
            refr = False

        if title2 in [pos.caption for pos in n._tabs._pages]:
            for id, pos in enumerate(n._tabs._pages):
                if pos.caption == title2:
                    n.SetSelection(id)
                    n.activate_page(pos.window)
            if refr:
                self._mgr.GetPane(panel).Show()
                self._mgr.Update()
            return None

        page = SchNotebookPage(n)
        if panel == "desktop2":
            if title is None:
                if type(address_or_parser) == str:
                    n.add_and_split(page, "", wx.RIGHT)
                else:
                    n.add_and_split(page, address_or_parser.title, wx.RIGHT)
            else:
                n.add_and_split(page, title2, wx.RIGHT)
        else:
            if title is None:
                if type(address_or_parser) == str:
                    n.AddPage(page, "", True)
                else:
                    n.AddPage(page, address_or_parser.title, True)
            else:
                n.AddPage(page, title2, True)

        if type(address_or_parser) == str:
            address = address_or_parser
        elif type(address_or_parser) == HttpResponse:
            address = address_or_parser.url

        else:
            address = address_or_parser.address

        if http:
            page.http = http
        else:
            page.http = wx.GetApp().get_http_for_adr(address)

        if refr:
            self._mgr.GetPane(panel).Show()
        self._mgr.Update()

        return page.new_child_page(address_or_parser, None, parameters)

    def on_taskbar_hide(self, event):
        self.Hide()

    def on_taskbar_toogle(self, event):
        if self.IsShown():
            self.Hide()
        else:
            self.Show()

    def on_taskbar_show(self, event):
        self.Show()

    def bind_command(self, fun, id=wx.ID_ANY):
        """bind command event, unlike wxPython Bind this function bind command to menu and toolbar interface"""
        self.Bind(wx.EVT_MENU, fun, id=id)
        if self.toolbar_interface:
            self.toolbar_interface.bind(fun, id)

    def get_menu_bar(self):
        """return toolbar interface"""
        return self.menubar_interface

    def get_tool_bar(self):
        """return toolbar interface"""
        return self.toolbar_interface

    def on_child_focus(self, event):
        pane = self._mgr.GetPane(event.GetWindow())
        if pane.IsOk():
            if self.active_pane:
                name = self.active_pane.Name
            else:
                name = ""
            if name != pane.Name:
                self.last_pane = self.active_pane
                self.active_pane = pane

        self.active_child_ctrl = event.GetWindow()
        event.Skip()

    def on_timer(self, evt):
        # if platform.system() == "Windows":
        #    wx.html2.WebView.New("messageloop")
        global _RECORD_VIDEO, _RECORD_VIDEO_STRUCT
        if self._video:
            if _RECORD_VIDEO == 0:
                try:
                    from mss import mss
                    import cv2
                    import numpy

                    _RECORD_VIDEO_STRUCT = (mss, cv2, numpy)
                    _RECORD_VIDEO = 1
                except:
                    _RECORD_VIDEO = 2
            if _RECORD_VIDEO == 1:
                save_video_frame(self)

        x = dispatcher.getReceivers(signal="PROCESS_INFO")
        if len(x) > 0:
            if not self._proc_mannager:
                self._proc_mannager = get_process_manager()
            x = self._proc_mannager.list_threads(all=False)
            dispatcher.send("PROCESS_INFO", self, x)

    def _append_command(self, typ, command):
        id = wx.NewId()
        self.command[id] = (typ, command)
        return id

    def _create_bars(self, bar, tab):
        for row in tab:
            if len(row[0].data) > 0:
                pos = 1
            else:
                if len(row[1].data) > 0:
                    pos = 2
                else:
                    if len(row[2].data) > 0:
                        pos = 3
                    else:
                        pos = -1
            if pos >= 0:
                if pos == 1:
                    page = bar.append(row[0].data)
                if pos == 2:
                    panel = page.append(row[1].data)
                if pos == 3:
                    try:
                        bitmap = (wx.GetApp().images)[int(row[4].data)]
                    except:
                        if row[4].data and row[4].data != "" and row[4].data != "None":
                            try:
                                bitmap = bitmap_from_href(row[4].data.split("<=")[0])
                            except:
                                bitmap = (wx.GetApp().images)[0]
                        else:
                            bitmap = (wx.GetApp().images)[0]
                    if not bitmap:
                        bitmap = (wx.GetApp().images)[0]
                    idn = self._append_command(row[5].data, row[6].data)
                    panel.append(idn, row[2].data, bitmap)
            bar.bind(self.on_command)

    def _create_menu_bar(self):
        tab = wx.GetApp().get_tab(self._menu_bar_lp)[1:]
        bar = self.menubar_interface
        return self._create_bars(bar, tab)

    def _create_tool_bar(self):
        tab = wx.GetApp().get_tab(self._toolbar_bar_lp)[1:]
        bar = self.toolbar_interface
        return self._create_bars(bar, tab)

    def bind_to_toolbar(self, funct, id):
        """bind function to toolbar button

        Args:
            funct - function (event handler) to bind
            id - toolbar button id
        """
        if "toolbar" in self.gui_style:
            self.toolbar_interface.bind(funct, id)

    def _create_status_bar(self):
        statusbar = self.CreateStatusBar(2)
        statusbar.SetStatusWidths([-2, -3])
        statusbar.SetStatusText("Ready", 0)
        return statusbar

    def _exit(self, event=None):
        finish_video()
        self.t1.Stop()
        wx.GetApp().on_exit()
        self._mgr.UnInit()
        if self.tbIcon:
            self.tbIcon.RemoveIcon()
            self.tbIcon = None
        self.Destroy()

    def on_open(self, event):
        if wx.GetApp().base_app:
            base_path = wx.GetApp().base_address + "/" + wx.GetApp().base_app
        else:
            base_path = wx.GetApp().base_address
        self.new_main_page(
            base_path + "/schcommander/form/FileManager/", "Commander", None, "panel"
        )

    def on_exit(self, event=None):
        self._exit()

    def on_close(self, event):
        super().on_close(event)
        self._exit()
        event.Skip()

    def on_show_elem(self, event):
        name = ["header", "panel", "footer", "tb1", "tb2"][
            event.GetId() - ID_SHOWHEADER
        ]
        panel = self._mgr.GetPane(name)
        panel.Show(not panel.IsShown())
        self._mgr.Update()

    def count_shown_panels(self, count_toolbars=True):
        """count visible panels"""
        count = 0
        for panel in self._mgr.GetAllPanes():
            if panel.IsShown():
                if (not "Toolbar" in panel.caption) or count_toolbars:
                    count += 1
        return count

    def on_show_status_bar(self, event):
        if not self.statusbar:
            self._create_status_bar()

    def _on_html(self, command):
        l = command.split(",")
        if len(l) > 2:
            parm = l[2]
            if parm == "":
                parm = None
        else:
            parm = None

        if l[1] != None and l[1][0] == " ":
            l[1] = (l[1])[1:]
        if parm != None and parm[0] == " ":
            parm = parm[1:]

        if l[0]:
            return self.new_main_page(l[1], l[0], parm)
        else:
            return self.new_main_page(l[1], l[0], parm, panel="pscript")

    def _on_python(self, command):
        exec(command)

    def _on_sys(self, command):
        (self.sys_command)[command]()

    def on_command(self, event):
        id = event.GetId()
        if id in self.command:
            cmd = self.command[id]
            if cmd[0] == "html":
                return self._on_html(cmd[1])
            if cmd[0] == "python":
                return self._on_python(cmd[1])
            if cmd[0] == "sys":
                return self._on_sys(cmd[1])
        else:
            if id == ID_RESET:
                from pytigon_gui.toolbar.moderntoolbar import RibbonInterface

                old_toolbar = self.toolbar_interface
                sizer = self.GetSizer()
                self.toolbar_interface = RibbonInterface(self, self.gui_style)
                self._create_tool_bar()
                self.toolbar_interface.realize_bar()
                sizer.Replace(old_toolbar.get_bar(), self.toolbar_interface.get_bar())
                self.toolbar_interface.get_bar().SetSize(
                    old_toolbar.get_bar().GetSize()
                )
                wx.CallAfter(old_toolbar.get_bar().Destroy)
                return
            elif id == ID_WEB_NEW_WINDOW:
                win = (
                    wx.GetApp()
                    .GetTopWindow()
                    .new_main_page("^standard/webview/widget_web.html", "Empty page")
                )
                win.body.new_child_page("^standard/webview/gotopanel.html", title="Go")
                return
        event.Skip()

    def on_about(self):
        msg = (
            "Pytigon runtime\nSławomir Chołaj\nslawomir.cholaj@gmail.com\n\n"
            + "The program uses wxpython library version:"
            + wx.VERSION_STRING
        )
        dlg = (
            wx.MessageDialog(self, msg, "Pytigon", wx.OK | wx.ICON_INFORMATION)
            % wx.GetApp().title
        )
        dlg.ShowModal()
        dlg.Destroy()

    def on_acc_key_down(self, event):
        for a in self.aTable:
            if event.KeyCode == a[1]:
                if event.AltDown() and (a[0] & wx.ACCEL_ALT == 0):
                    continue
                if event.ControlDown() and (a[0] & wx.ACCEL_CTRL == 0):
                    continue
                if event.ShiftDown() and (a[0] & wx.ACCEL_SHIFT == 0):
                    continue
                self.ProcessEvent(wx.CommandEvent(wx.wxEVT_COMMAND_MENU_SELECTED, a[2]))
                return
        event.Skip()

    # def get_start_position(self):
    #    self.x = self.x + 20
    #    x = self.x
    #    pt = self.ClientToScreen(wx.Point(0, 0))
    #    return wx.Point(pt.x + x, pt.y + x)

    def goto_panel(self, panel_name):
        """Activate panel

        Args:
            panel_name: name of panel to activate
        """
        panel = self._mgr.GetPane(panel_name)
        if panel.window.GetPageCount() > 0:
            if not panel.IsShown():
                panel.Show(True)
                self._mgr.Update()
            panel.window.SetFocus()

    def on_goto_panel(self, event):
        panel = self._mgr.GetPane("panel")
        if panel.IsShown():
            self.goto_panel("panel")
        else:
            panel = self._mgr.GetPane("menu")
            if panel and panel.IsShown():
                panel.window.SetFocus()

    def on_goto_head(self, event):
        self.goto_panel("header")

    def on_goto_footer(self, event):
        self.goto_panel("footer")

    def on_goto_desktop(self, event):
        self.desktop.SetFocus()

    def show_pdf(self, response, title="", parameters=None):
        """show pdf downloaded from web server

        Args:
            page: web page address
        """

        temp_filename = schfs.get_temp_filename(ext="pdf")
        with schfs.open_file(temp_filename, "wb") as f:
            f.write(response.ptr())

        form_frame = self.new_main_page(
            "file://" + temp_filename, temp_filename, view_in="browser"
        )

        def _after_init():
            form_frame.body.WEB.execute_javascript("document.title = '%s';" % title)

        wx.CallAfter(_after_init)

        return form_frame

    def show_spdf(self, response, title="", parameters=None):
        temp_filename = schfs.get_temp_filename(ext="spdf", for_vfs=False)
        with schfs.open_file(temp_filename, "wb", for_vfs=False) as f:
            f.write(response.ptr())

        return self.new_main_page(
            "^standard/html_print/html_print.html",
            "file://" + temp_filename,
            parameters=temp_filename,
        )

    def print_spdf(self, response, title="", parameters=None):
        if hasattr(wx.GetApp(), "get_printout"):
            temp_filename = schfs.get_temp_filename(ext="spdf", for_vfs=False)
            with schfs.open_file(temp_filename, "wb", for_vfs=False) as f:
                f.write(response.ptr())

            printer = wx.Printer()
            printout = wx.GetApp().get_printout(temp_filename)
            printer.Print(self, printout, True)

            os.unlink(temp_filename)
        return

    def show_document(self, response, title="", parameters=None):
        cd = response.response.headers.get("content-disposition")
        if cd:
            name = cd.split("filename=")[1].replace('"', "")
        else:
            name = "data.dat"

        p = response.ptr()
        path = gettempdir()

        file_path = os.path.join(path, name)

        with open(file_path, "wb") as f:
            f.write(p)

        if platform.system() == "Windows":
            os.startfile(file_path)
        elif platform.system() == "Darwin":
            subprocess.Popen(["open", file_path])
        else:
            subprocess.Popen(["xdg-open", file_path])

    def show_image(self, response, title="", parameters=None):
        temp_filename = schfs.get_temp_filename(ext="spdf", for_vfs=False)
        with schfs.open_file(temp_filename, "wb", for_vfs=False) as f:
            f.write(response.ptr())

        return self.new_main_page(
            "^standard/image_viewer/viewer.html", temp_filename, view_in="desktop"
        )

    def download_data(self, response, title="", parameters=None):
        cd = response.response.headers.get("content-disposition")
        if cd:
            name = cd.split("filename=")[1].replace('"', "")
        else:
            name = "data.dat"

        p = response.ptr()

        path = str(Path.home() / "Downloads")

        with open(os.path.join(path, name), "wb") as f:
            f.write(p)

        if platform.system() == "Windows":
            os.startfile(path)
        elif platform.system() == "Darwin":
            subprocess.Popen(["open", path])
        else:
            subprocess.Popen(["xdg-open", path])

        return True

    def get_main_panel(self):
        return self._panel



"""Module contain SchForm class."""

import gc
import sys
import types
import math

import wx
from wx.lib.scrolledpanel import ScrolledPanel
import wx.lib.agw.ribbon as RB

from pytigon_lib.schtools import createparm
from pytigon_lib.schparser.html_parsers import ShtmlParser
from pytigon_lib.schtools.tools import clean_href, is_null
from pytigon_lib.schhtml.wxdc import DcDc
from pytigon_lib.schhtml.htmlviewer import HtmlViewerParser

_ = wx.GetTranslation

_INIT_CSS_STR = None
_PRE_PRECESS_LIB = []


def install_pre_process_lib(fun):
    global _PRE_PRECESS_LIB
    _PRE_PRECESS_LIB.append(fun)


def _get_css():
    global _INIT_CSS_STR
    if _INIT_CSS_STR == None:
        with open(wx.GetApp().src_path + "/appdata/icss/_core.icss", "r") as f:
            _INIT_CSS_STR = f.read()
        with open(wx.GetApp().src_path + "/appdata/icss/wx.icss", "r") as f:
            _INIT_CSS_STR += f.read()
    return _INIT_CSS_STR


class SchForm(ScrolledPanel):
    """Html window"""

    @property
    def bestsize(self):
        return self._bestsize

    @bestsize.setter
    def bestsize(self, x):
        self._bestsize = x

    def __init__(
        self, parent, page=None, hscroll=False, vscroll=False, form_type="body"
    ):
        """Constructor

        Args:
            parent - parent window
            page - parent SchPage object
            hscroll - enable horizontal scroll bar
            vscroll - enable vertical scroll  bar
            form_type - can be: body, header, footer, panel
        """
        self._bestsize = None
        self._scroll_xy = (0, 0)
        self._act_scroll_xy = (0, 0)
        self._enabled_controls = False
        self._last_size = (-1, -1)
        self._best_virtual_size = None
        self._block = False
        self._dc_buf = None
        self._dc_buf_x = 0
        self._dc_buf_y = 0
        self._lock = False

        self.no_vscrollbar = not vscroll
        self.no_hscrollbar = not hscroll
        self.address = None
        self.script = None
        self.parameters = None
        self.page_source = "<html></html>"
        self.script = None
        self.bestsize = None
        self.hover_obj = None
        self.cursor_type = 0
        self.wxdc = None
        self.after_init = False
        self.update_controls = False
        self.obj_action_dict = {}
        self.obj_id_dict = {}
        self.last_control_with_focus = None
        self.page = page
        self.form_type = form_type
        self.init_css_str = None
        self.evt_ind = -1
        self.closing = False
        self.acc_tabs = {}
        self.acc_tab = None
        self.acc_tab_bind = False
        self.last_clicked = None
        self.t1 = None
        self.websockets = []

        self.hscroll = hscroll
        self.vscroll = vscroll

        ScrolledPanel.__init__(self, parent, -1, style=wx.WANTS_CHARS)
        try:
            self.SetBackgroundStyle(wx.BG_STYLE_ERASE)
        except:
            self.SetBackgroundStyle(wx.BG_STYLE_CUSTOM)

        self.SetupScrolling(hscroll, vscroll, rate_y=1)

        self.SetAcceleratorTable(wx.AcceleratorTable(wx.GetApp().GetTopWindow().aTable))

        self.Bind(wx.EVT_SIZE, self.on_size)
        self.Bind(wx.EVT_LEFT_DOWN, self.on_left_down)
        self.Bind(wx.EVT_LEFT_UP, self.on_left_up)
        self.Bind(wx.EVT_RIGHT_UP, self.on_right_up)
        self.Bind(wx.EVT_MOTION, self.on_motion)
        self.Bind(wx.EVT_ERASE_BACKGROUND, self.on_erase_background)
        self.Bind(wx.EVT_SET_FOCUS, self.on_focus)

        self.EnableScrolling(False, False)

    def SetFocus(self):
        if self.last_control_with_focus:
            self.last_control_with_focus.SetFocus()
        else:
            wx.CallAfter(self.Navigate, None)

    def Navigate(self, ctrl, back=False):
        next = False
        children = [
            child for child in list(self.GetChildren()) if child.CanAcceptFocus()
        ]
        if back:
            widgets = reversed(children)
        else:
            widgets = children
        for w in widgets:
            if next or not ctrl:
                w.SetFocus()
                self.last_control_with_focus = w
                return
            if w == ctrl:
                next = True
        if children and len(children) > 0:
            if back:
                children[-1].SetFocus()
                self.last_control_with_focus = children[-1]
            else:
                children[0].SetFocus()
                self.last_control_with_focus = children[0]

    def on_focus(self, event):
        if self.last_control_with_focus:
            self.last_control_with_focus.SetFocus()
        else:
            wx.CallAfter(self.Navigate, None)

    def bind_to_ctrl(self, ctrl, id, fun, fun2=None):
        """bind function callback to command event from ctrl

        Args:
            ctrl - widget
            id - command id
            fun - callback function command event
            fun2 - callback function for EVT_UPDATE_UI event

        """
        if fun2:
            ctrl.Bind(wx.EVT_UPDATE_UI, fun2, id=id)
        ctrl.Bind(RB.EVT_RIBBONBUTTONBAR_CLICKED, fun, id=id)
        ctrl.Bind(RB.EVT_RIBBONTOOLBAR_CLICKED, fun, id=id)
        ctrl.Bind(wx.EVT_MENU, fun, id=id)

    def scroll_child_into_view(self, child):
        """scroll this window to position at which child window is visible"""
        if child.GetParent() == self:
            (sppu_x, sppu_y) = self.GetScrollPixelsPerUnit()
            (vs_x, vs_y) = self.GetViewStart()
            cr = child.GetRect()
            clntsz = self.GetClientSize()
            (new_vs_x, new_vs_y) = (-1, -1)
            if cr.y < 0 and sppu_y > 0:
                new_vs_y = vs_y + cr.y / sppu_y
            if cr.bottom > clntsz.height and sppu_y > 0:
                diff = math.ceil((1.0 * (cr.bottom - clntsz.height + 1)) / sppu_y)
                if cr.y - diff * sppu_y > 0:
                    new_vs_y = vs_y + diff
                else:
                    new_vs_y = vs_y + cr.y / sppu_y
            if new_vs_y != -1:
                self.Scroll(-1, new_vs_y)

    def set_css(self, css_str):
        """set css file for this form"""
        self.init_css_str = css_str

    def get_css(self):
        if self.init_css_str:
            return self.init_css_str
        else:
            return _get_css()

    def GetBestVirtualSize(self):
        if self._best_virtual_size:
            return self._best_virtual_size
        else:
            try:
                (dx, dy) = self.GetSize()
            except:
                return (0, 0)
            # dx -= wx.SystemSettings.GetMetric(wx.SYS_VSCROLL_X) + 1
            # dy -= wx.SystemSettings.GetMetric(wx.SYS_HSCROLL_Y) + 1
            return (int(dx), int(dy))

    def _calculate_size(self, width):
        dc = wx.ClientDC(self)
        wxdc = DcDc(dc, calc_only=True, width=width, height=-1)
        p = HtmlViewerParser(
            dc=wxdc, calc_only=True, init_css_str=self.get_css(), css_type=1
        )
        p.set_http_object(wx.GetApp().http)
        p.set_parent_window(self)
        if not self._lock:
            self._lock = True
            self.GetParent().restart_ctrl_lp()
            p.feed(self.page_source)
            self.GetParent().remove_old_ctrls()
            self._lock = False
        (dx, dy) = p.get_max_sizes()
        p.close()
        return (dx, dy)

    def calculate_best_size(self):
        psize = self.GetParent().GetParent().GetClientSize()
        if self.bestsize:
            (dx, dy) = self._calculate_size(self.bestsize[0])
        else:
            (dx, dy) = self._calculate_size((psize[0] * 3) / 4)
        self.bestsize = (dx, dy + 8)
        return self.bestsize

    def html_updatet(self, dc, size):
        """parse a html page after it was updated"""
        w = size.GetWidth()
        h = size.GetHeight()
        if w == 20 and h == 20:
            return
        if not self.wxdc:
            if self.no_vscrollbar:
                self.wxdc = DcDc(
                    dc, calc_only=False, width=w - 1, height=-1, record=True
                )
            else:
                self.wxdc = DcDc(
                    dc,
                    calc_only=False,
                    width=w - 1 - wx.SystemSettings.GetMetric(wx.SYS_VSCROLL_X),
                    height=-1,
                    record=True,
                )
            p = HtmlViewerParser(
                dc=self.wxdc, calc_only=False, init_css_str=self.get_css(), css_type=1
            )
            p.set_http_object(wx.GetApp().http)
            p.set_parent_window(self)
            if not self._lock:
                self._lock = True
                self.GetParent().restart_ctrl_lp()
                p.feed(self.page_source)
                self.GetParent().remove_old_ctrls()
                self._lock = False
            self.obj_action_dict = p.obj_action_dict
            self.obj_id_dict = p.obj_id_dict
            if not self.no_vscrollbar:
                (w2, h2) = p.get_max_sizes()
                # h2+=8
                # h2-=wx.SystemSettings.GetMetric(wx.SYS_VSCROLL_X)
                self.hscroll = True if w2 > w + 3 else False
                self.vscroll = True if h2 > h + 3 else False
                if self.no_hscrollbar:
                    self.hscroll = False
                if (
                    (self.vscroll or self.hscroll)
                    and w > 0
                    and h > 0
                    and w2 > 0
                    and h2 > 0
                ):
                    if self.vscroll:
                        w2 = w2 + wx.SystemSettings.GetMetric(wx.SYS_VSCROLL_X)
                    if self.hscroll:
                        h2 = h2 + wx.SystemSettings.GetMetric(wx.SYS_HSCROLL_Y)
                    self._best_virtual_size = (int(w2), int(h2))
                    self.EnableScrolling(self.hscroll, self.vscroll)
                    self._SetupAfter(False)
                self.SetVirtualSize((w2, h2))
            else:
                self.EnableScrolling(False, False)
                self.SetVirtualSize((w, h))
            p.close()
            self.update_controls = False
            wx.CallAfter(self.SetupScrolling, self.hscroll, self.vscroll, rate_y=1)

    def draw_background(self, refresh_all=False, size=None):
        """drawn a rendered html page as a background"""
        if self._block:
            return
        self._block = True
        if not self.wxdc:
            self.Scroll(0, 0)
        (xv, yv) = self.GetViewStart()
        (dx, dy) = self.GetScrollPixelsPerUnit()
        (x, y) = (xv * dx, yv * dy)
        self._act_scroll_xy = (x, y)
        dc = wx.ClientDC(self)
        if self.wxdc and self._dc_buf and self._dc_buf_x == x and self._dc_buf_y == y:
            dc.SetDeviceOrigin(-1 * x, -1 * y)
            rect = self.GetRect()
            dc.Blit(x, y, rect.GetWidth(), rect.GetHeight(), self._dc_buf, x, y)
        else:
            rect = self.GetRect()
            if rect.GetWidth() < 1 or rect.GetHeight() < 1:
                rect = wx.Rect(0, 0, 1, 1)  # width=1, height=1)
            dc.SetDeviceOrigin(-1 * x, -1 * y)
            bitmap = wx.Bitmap(rect.GetWidth(), rect.GetHeight())
            dc2 = wx.MemoryDC(bitmap)
            dc2.SetBackground(
                wx.Brush(wx.SystemSettings.GetColour(wx.SYS_COLOUR_3DFACE))
            )
            dc2.Clear()
            dc2.SetDeviceOrigin(-1 * x, -1 * y)
            if not self.wxdc:
                if size:
                    self.html_updatet(dc2, size)
                else:
                    self.html_updatet(dc2, rect)
            else:
                self.wxdc.dc = dc2
                self.wxdc.play()
            dc.Blit(x, y, rect.GetWidth(), rect.GetHeight(), dc2, x, y)
            self._dc_buf = dc2
            self._dc_buf_x = x
            self._dc_buf_y = y
        self._block = False

    def on_erase_background(self, event):
        self.draw_background()

    def append_ctrl(self, ctrl):
        try:
            setattr(self, ctrl.unique_name, ctrl)
        except:
            pass
        self.GetParent().append_ctrl(ctrl)

    def enable_ctrls(self, ctrls):
        """set widgets which are enabled or disabled
        Widget enabled is updated when html page is refreshed

        Args:
            ctrls - list or tuple with widgets names
        """
        self._enabled_controls = ctrls

    def is_ctrl_block(self, ctrl):
        if self._enabled_controls:
            if ctrl in self._enabled_controls:
                return False
            else:
                return True
        else:
            return True

    def restore_scroll_pos(self):
        wx.ScrolledWindow.Scroll(self, self._scroll_xy[0], self._scroll_xy[1])

    def enable(self, enable=True):
        """enable od disable the form"""
        ret = super(SchForm, self).Enable(enable)
        if enable:
            wx.CallAfter(self.restore_scroll_pos)
        return ret

    def set_best_size(self, bestsize):
        """Set the best size for this form"""
        self.bestsize = bestsize

    # def on_opening_url(self, type, url):
    #    if hasattr(self, 'filter_url'):
    #        f = self.filter_url(type, url)
    #        if f != None:
    #            return f
    #    if url.startswith('/'):
    #        url2 = url
    #        base = wx.GetApp().base_address
    #        if base.startswith('http://127.0.0.2'):
    #            if url.startswith('/app_media'):
    #                if ':' in wx.GetApp().root_path:
    #                    url2 = wx.GetApp().root_path + '/prj' + url.replace('/app_media', '')
    #                    url2 = norm_path(url2)
    #                else:
    #                    url2 = 'file://' + wx.GetApp().root_path + '/prj' + url.replace('/app_media', '')
    #                    url2 = norm_path(url2)
    #        else:
    #            url2 = base + url
    #        return url2
    #    return True

    def on_left_down(self, evt):
        parent = self.GetParent()
        while parent != None:
            if parent.__class__.__name__ == "HtmlPanel":
                parent.SelectTab()
                break
            parent = parent.GetParent()
        evt.Skip()

    def on_link_clicked(self, evt):
        href = evt.GetHref()
        target = evt.GetTarget()
        if target == "":
            target = "_blank"
        if not (href.startswith("http://") or href.startswith("/")):
            href = self.address.split("?")[0] + href
        wx.CallAfter(self.href_clicked, self, {"href": href, "target": target})
        return

    def refresh_time(self, number_of_sec):
        """When refresh time is set form automatically refresh his content every number_of_sec seconds"""
        self.t1 = wx.Timer(self)
        self.t1.Start(int(number_of_sec * 1000))
        self.Bind(wx.EVT_TIMER, self.on_timer, self.t1)

    # def close_with_delay(self):
    #    wx.CallAfter(self.cancel, True)

    def on_timer(self, event):
        if self.page:
            if not self.page.exists:
                self.t1.Stop()
                return
        self.GetParent().refresh_html()

    def cancel(self):
        """Close this form without saving its content"""
        if self.page:
            self.page.exists = False
        if self.t1:
            self.t1.Stop()
        if self.page:
            self.page.GetParent().on_child_form_cancel()

    def ok(self):
        if self.page:
            self.page.exists = False
        if self.t1:
            self.t1.Stop()
        if self.page:
            self.page.GetParent().on_child_form_ok()

    def set_htm_type(self, form_type):
        self.form_type = form_type

    def get_gparent_page(self):
        """return grand parent SChPage object in hierarchy"""
        page = self.get_parent_page()
        if page:
            return page.get_parent_page()
        else:
            return None

    def get_parent_page(self):
        """return parent SChPage object in hierarchy"""
        return self.page

    def get_parent_form(self):
        """return parent SchFrame object in hierarchy"""
        gpage = self.get_gparent_page()
        if gpage:
            return gpage.active_form
        return None

    def on_right_up(self, evt):
        okno = self.new_main_page(
            "^standard/editor/editor.html",
            self.get_parent_page().title + " - " + _("page source"),
            None,
        )

        def init_ctrl():
            okno.body.EDITOR.SetValue(self.page_source.tostream().getvalue())
            okno.body.EDITOR.GotoPos(0)

        wx.CallAfter(init_ctrl)

    def _get_obj_for_redraw(self, pos, type=0):
        if "href" in self.obj_action_dict:
            for obj in self.obj_action_dict["href"]:
                if type == 1 or obj.can_hover():
                    for r in obj.rendered_rects:
                        rect = wx.Rect(int(r[0]), int(r[1]), int(r[2]), int(r[3]))
                        if rect.Contains(pos):
                            return obj
        if type == 1:
            return None
        if "class" in self.obj_action_dict:
            for obj in self.obj_action_dict["class"]:
                if obj.can_hover():
                    for r in obj.rendered_rects:
                        rect = wx.Rect(r[0], r[1], r[2], r[3])
                        if rect.Contains(pos):
                            return obj
        for id in self.obj_id_dict:
            obj = self.obj_id_dict[id]
            for r in obj.rendered_rects:
                if obj.can_hover():
                    rect = wx.Rect(r[0], r[1], r[2], r[3])
                    if rect.Contains(pos):
                        return obj
        return None

    def on_motion(self, evt):
        pos = evt.GetPosition()
        pos2 = (pos[0] + self._act_scroll_xy[0], pos[1] + self._act_scroll_xy[1])
        self.redraw_html_elems(pos2)
        evt.Skip()

    def redraw_html_elems(self, pos):
        obj = self._get_obj_for_redraw(pos)
        if obj or self.hover_obj:
            dc = wx.ClientDC(self)
            self.wxdc.dc = dc
            if self.hover_obj:
                if obj != self.hover_obj:
                    self.hover_obj.gparent.set_hover(False)
                    self.hover_obj.gparent.refresh()
            if obj:
                if obj != self.hover_obj and obj.can_hover():
                    obj.gparent.set_hover(True)
                    obj.gparent.refresh()
        self.hover_obj = obj
        if self._get_obj_for_redraw(pos, type=1):
            cursor_type = 1
        else:
            cursor_type = 0
        if cursor_type != self.cursor_type:
            if cursor_type == 0:
                self.SetCursor(wx.Cursor(wx.CURSOR_ARROW))
            elif cursor_type == 1:
                self.SetCursor(wx.Cursor(wx.CURSOR_HAND))
            else:
                self.SetCursor(wx.Cursor(wx.CURSOR_ARROW))
            self.cursor_type = cursor_type

    def on_left_up(self, evt):
        pos = evt.GetPosition()
        pos2 = (pos[0] + self._act_scroll_xy[0], pos[1] + self._act_scroll_xy[1])
        if "href" in self.obj_action_dict:
            for obj in self.obj_action_dict["href"]:
                for r in obj.rendered_rects:
                    rect = wx.Rect(int(r[0]), int(r[1]), int(r[2]), int(r[3]))
                    if rect.Contains(pos2):
                        self.href_clicked(self, obj.attrs)
                        evt.Skip()
                        return
        evt.Skip()

    def CanClose(self):
        """Test if form can be closed"""
        ret = True
        if hasattr(self, "can_close"):
            ret = self.can_close()
        else:
            widgets = list(self.GetParent().get_widgets().values())
            for w in widgets:
                if hasattr(w, "CanClose"):
                    if not w.CanClose():
                        ret = False
        return ret

    def _on_close(self):
        app = wx.GetApp()
        self.closing = True
        if hasattr(self, "on_close"):
            self.on_close()
        for x in self.websockets:
            app.remove_websocket_callback(x[0], x[1])
        gc.collect()

    def any_parent_command(self, command, *args, **kwds):
        """find first in hierarchy window which contains function with specyfied name and run this function

        Args:
            command: name of function to find
            args, kwds: parameters forwarded to function
        """
        parent = self.GetParent()
        while parent != None:
            if hasattr(parent, command):
                return getattr(parent, command)(*args, **kwds)
            parent = parent.GetParent()
        print("METHOD NOT FOUND!: ", command)
        return None

    def set_address_parm(self, address):
        """page address for refr_obj in format http://adres?zmienna1|zmienna2,zmienna"""
        if address:
            elementy = address.split("|")
            if elementy[0][-1] == "/":
                address2 = address
            else:
                address2 = elementy[0] + "/"
                if len(elementy) > 1:
                    address2 = address2 + "|" + elementy[1]
            self.address = address2

    def get_parm_obj(self):
        """Get parent object width method get_par"""
        parent = self
        while parent != None:
            if hasattr(parent, "get_parm"):
                return parent
            parent = parent.GetParent()
        return None

    def refr(self, refr_always=False):
        # """Refresh form"""
        # self.refr_obj(refr_always=True)
        # def refr_obj(self, refr_always=False):
        """Refresh html content"""

        if refr_always or self.GetParent().IsShown():
            if self.address:
                parm = createparm.create_parm(self.address, self.get_parm_obj())
                http = wx.GetApp().get_http(self)
                if parm:
                    response = http.get(self, str(parm[0] + parm[1] + parm[2]))
                else:
                    response = http.get(self, self.address)
                if response.ret_code == 404:
                    return

                mp = ShtmlParser()
                mp.process(response.str(), "local")
                if self.form_type == "panel":
                    strona = mp.get_panel()
                elif self.form_type == "header":
                    strona = mp.get_header()
                elif self.form_type == "footer":
                    strona = mp.get_footer()
                else:
                    strona = mp.get_body()
                if self.show_form(strona, self.parameters):
                    self.init()

    def on_size(self, event):
        if event.GetSize() == (0, 0) or event.GetSize() == (20, 20):
            self._last_size = event.GetSize()
            event.Skip()
            return
        if self._last_size != event.GetSize():
            self._last_size = event.GetSize()
            self.wxdc = None
            self.draw_background(True, event.GetSize())
            event.Skip()

    def set_page(self, page_source):
        """Set html page source"""
        self.page_source = page_source
        self.wxdc = None
        self.draw_background()

    def pre_process_page(self, page):
        global _PRE_PRECESS_LIB
        page2 = page
        for fun in _PRE_PRECESS_LIB:
            page2 = fun(self, page2)
        return page2

    def exec_code(self, script, parm=None):
        """run code attached to html page"""
        if not parm:
            d = {"wx": wx, "self": self}
        else:
            d = parm
            d["self"] = self
        try:
            exec(script, d)
        except:
            print("### ERROR IN SCRIPT ###################################")
            print(script)
            print("#######################################################")
            import traceback

            print(sys.exc_info()[0])
            print(traceback.print_exc())
            print("#######################################################")

    def show_form(self, page_and_script, parameters=None):
        """Show form

        Args:
            page_and_script: list or touple, 0: html page content, 1: script
            parameters: parameters forwarded to http server in request
        """
        if self.page and self.page.exists == False:
            return False

        self.page_source = self.pre_process_page(page_and_script[0])
        self.parameters = parameters
        if not self.script:
            self.script = str(page_and_script[1])
            if self.script and len(self.script) > 1:
                app = wx.GetApp()
                main_window = app.GetTopWindow()
                desktop = (main_window.desktop,)
                mgr = main_window._mgr
                menu_bar = main_window.get_menu_bar()
                script_tmp = self.script.replace("\r", "").split("\n")
                script = ""
                tab = -1
                for line in script_tmp:
                    if tab == -1:
                        if line.strip() != "":
                            tab = len(line) - len(line.lstrip())
                    if tab >= 0:
                        script = script + line[tab:] + "\n"
                d = {"wx": wx}
                self.exec_code(script, d)

                for key in d:
                    if not key.startswith("__"):
                        if hasattr(d[key], "__call__"):
                            fun = d[key]
                            try:
                                method = types.MethodType(d[key], self, self.__class__)
                            except:
                                method = types.MethodType(d[key], self)
                            if hasattr(self, key):
                                setattr(self, "base_" + key, getattr(self, key))
                            setattr(self, key, method)
                        else:
                            setattr(self, key, d[key])
        return True

    def init(self, callback=None):
        """init form"""
        if not self.after_init:
            if hasattr(self, "init_form"):

                def _init():
                    nonlocal self
                    self.init_form()
                    if callback:
                        callback(self)

                wx.CallAfter(_init)
            self.after_init = True
        else:
            if hasattr(self, "reinit"):
                self.reinit()
            elif hasattr(self, "init_form"):
                self.init_form()
            if callback:
                callback(self)
        wx.CallAfter(self._build_acc_tab)
        wx.CallAfter(self._check_scroll_bar)

    def _check_scroll_bar(self):
        size = self.GetSize()
        max_y = 0
        children = list(self.GetChildren())
        for child in children:
            pos = child.GetPosition()
            child_size = child.GetSize()
            y = pos[1] + child_size[1]
            if y > size[1]:
                if y > max_y:
                    max_y = y
        if max_y > 0:
            self.vscroll = True
            self.no_vscrollbar = False
            self.EnableScrolling(self.hscroll, self.vscroll)
            self.SetVirtualSize(size[0], max_y)
            wx.CallAfter(self.SetupScrolling, self.hscroll, self.vscroll, rate_y=1)
            self.wxdc = None
            self.draw_background(True, size)
            self.Refresh()
            return

    def new_local_child_page(self, address, title="", parameters=None):
        return self.new_child_page("http://local.net/" + address, title, parameters)

    def new_plugin_child_page(self, path, address, title="", parameters=None):
        p = path.split("/")
        address2 = "appdata/plugins/" + p[-3] + "/" + p[-2] + "/" + address
        return self.new_local_child_page(address2, title, parameters)

    def new_child_page(
        self, address_or_parser, title="", parameters=None, callback=None
    ):
        """Create a child page

        Args:
            address_or_parser: can be: address of http page (str type) or
            :class:'~pytigon_lib.schparser.html_parsers.ShtmlParser'
            title - new page title
            parameters: dict
        """
        self._scroll_xy = self.GetViewStart()
        self.get_parm_obj().active_form = self
        self.GetParent().active_form = self
        if hasattr(self.GetParent(), "last_control_with_focus"):
            self.last_control_with_focus = self.GetParent().last_control_with_focus
        else:
            self.last_control_with_focus = None
        self.GetParent().disable_setfocus = True
        return self.any_parent_command(
            "new_child_page", address_or_parser, title, parameters, callback=callback
        )

    def new_main_page(
        self, address_or_parser, title=None, parameters=None, view_in=None
    ):
        return self.any_parent_command(
            "new_main_page", address_or_parser, title, parameters, view_in
        )

    def scroll_to_href(self, href):
        """scroll this window to position at which widget with a href attribute is visible"""
        ref = href[1:]
        if ref in self.obj_id_dict:
            obj = self.obj_id_dict[ref]
            rects = obj.rendered_rects
            if len(rects) > 0:
                rect = rects[0]
                x = rect[0]
                y = rect[1]
                if self.GetScrollPixelsPerUnit()[0] > 0:
                    dx = x / self.GetScrollPixelsPerUnit()[0]
                else:
                    dx = -1
                if self.GetScrollPixelsPerUnit()[1] > 0:
                    dy = y / self.GetScrollPixelsPerUnit()[1]
                else:
                    dy = -1
                self.Scroll(dx, dy)

    def href_clicked(self, ctrl, attr_dict, upload=False, fields=False):
        """Handle action connected to widget with href attribute

        Args:
            ctrl - widget
            attr_dict - dict with attributes for http request
            upload - if true, this is action prepared to upload content to server
            fields - list of fields which values are sended to http server
        """
        self.last_clicked = ctrl

        if "href" in attr_dict:
            href = attr_dict["href"]
            if href and len(href) > 0 and href[0] == "#":
                return self.scroll_to_href(href)
            if href:
                href = clean_href(href)
            if "target" in attr_dict:
                target = attr_dict["target"]
            else:
                target = "_blank"
            if "title" in attr_dict:
                title = attr_dict["title"]
            else:
                title = None

            if href and hasattr(self, "filter_url"):
                f = self.filter_url(target, href)
                if f != None:
                    if type(f) == str:
                        href = clean_href(f)
                    else:
                        return f

            if href:
                parent_form = self.get_parent_form()
                if hasattr(parent_form, "filter_child_url"):
                    ret = parent_form.filter_child_url(self, href)
                    if ret:
                        return ret
                # parent_page = self.GetParent().get_parent_page()
                # if hasattr(parent_page.body, "filter_child_url"):
                #    ret = parent_page.body.filter_child_url(self, href)
                #    if ret:
                #        return ret

            if href == None:
                if target == "_parent":
                    self.any_parent_command("on_child_form_cancel")
                    return
                if target == "_refresh":
                    self.any_parent_command("send_refr_obj")
                if target == "refresh_frame":
                    self.any_parent_command("_refresh")
                return
            if href[:7] == "http://" or href[:7] == "file://":
                adr = href
            else:
                if href:
                    if href[0] == ".":
                        elm1 = self.get_parm_obj().address.split("?")[0]
                        if not elm1[-1] == "/":
                            id = elm1.rfind("/")
                            if id >= 0:
                                elm1 = elm1[: id + 1]
                            else:
                                elm1 = ""
                        adr = elm1 + href
                    else:
                        adr = href
                else:
                    adr = self.get_parm_obj().address

            if "/pdf/view" in adr:
                wx.GetApp().GetTopWindow().show_pdf(adr)
                return

            if "/odf/view" in adr:
                wx.GetApp().GetTopWindow().show_odf(adr)
                return

            parm = createparm.create_parm(adr, self.get_parm_obj(), no_encode=True)
            if parm:
                adr = parm[0]

            if fields:
                (typ, fields2) = fields.split(":")
                adr = adr + "|" + fields2
            if target == "refresh_frame":
                self.any_parent_command("set_adr_and_param", adr, self.get_parm_obj())
                self.any_parent_command("refresh_html")
                return
            parm = createparm.create_parm(adr, self.get_parm_obj(), no_encode=True)
            post = None
            if parm:
                if typ in ("POST", "post"):
                    adr = parm[0]
                    post = parm[2]
                else:
                    adr = parm[0] + parm[1] + parm[2]
                    post = None

            http = wx.GetApp().get_http(ctrl)

            if upload:
                response = http.post(self, adr, post, upload=True)
            else:
                if post:
                    response = http.post(self, adr, post)
                else:
                    response = http.get(self, adr)

            if "500" in response.ret_content_type:
                return

            if "text/plain" in response.ret_content_type:
                s = response.str()
                if title:
                    edit_name = title
                else:
                    edit_name = href
                okno = self.new_main_page("^standard/editor/editor.html", href, None)
                okno.body.EDITOR.SetValue(s)
                okno.body.EDITOR.GotoPos(0)
            elif "/json" in response.ret_content_type:
                x = response.json()
                if "action" in x and x["action"] in (
                    "update_row_ok",
                    "new_row_ok",
                    "delete_row_ok",
                ):
                    if "obj" in x:
                        obj = x["obj"]
                    else:
                        obj = None
                    if self.GetParent().parent_page:
                        ret = self.GetParent().parent_page.signal(x["action"], obj)
                    else:
                        ret = self.GetParent().signal(x["action"], obj)
                    if ret:
                        self.cancel()
                    else:
                        self.ok()
                return
            else:
                if "text/" not in response.ret_content_type:
                    wx.GetApp().GetTopWindow()._open_page(response)
                    return

                s = response.str()
                mp = ShtmlParser()
                mp.process(s, response.new_url)
                if "target" in mp.var:
                    target = mp.var["target"]

                if href and hasattr(self, "filter_http_result"):
                    f = self.filter_http_result(target, href, mp)
                    if f != None:
                        return f
                if target == "_self":
                    update_controls = self.update_controls
                    self.update_controls = True
                    self.show_form(mp.get_body(), self.get_parm_obj())
                    self.wxdc = None
                    self.draw_background()
                    self.Refresh()
                    self.init()
                    self.update_controls = update_controls
                    if "href" in attr_dict:
                        href = attr_dict["href"]
                    else:
                        href = None

                    if (
                        self.GetParent().address_or_parser.__class__.__name__
                        == "ShtmlParser"
                    ):
                        self.GetParent().address_or_parser.address = response.new_url
                    else:
                        self.GetParent().address_or_parser = response.new_url

                    return
                if (
                    target.startswith("popup")
                    or target.startswith("inline")
                    or target
                    in (
                        "_blank",
                        "_self",
                    )
                ):
                    self.GetParent().active_ctrl = ctrl
                    win = self.new_child_page(
                        mp, is_null(mp.title, title), parameters=self.get_parm_obj()
                    )
                    if win:
                        win.body.set_address_parm(str(response.new_url))
                    return
                if target == "_parent":
                    self.new_main_page(
                        mp,
                        is_null(mp.title, title),
                        parameters=self.get_parm_obj(),
                        view_in=None,
                    )
                    return
                if target == "refresh_page":
                    # self.any_parent_command("set_new_href", adr)
                    self.any_parent_command("refresh_html")

                if target.startswith("_top2"):
                    x = target[5:]
                    if x[0:1] == "_":
                        self.new_main_page(
                            mp,
                            is_null(mp.title, title),
                            parameters=self.get_parm_obj(),
                            view_in=x[1:],
                        )
                    else:
                        self.new_main_page(
                            mp,
                            is_null(mp.title, title),
                            parameters=self.get_parm_obj(),
                            view_in="desktop2",
                        )
                    return
                if target == "_parent":
                    self.any_parent_command("on_child_form_cancel")
                    return
                if target == "_refresh":
                    self.GetParent().refresh_html()
                    return
                if target == "_parent_refr":
                    self.any_parent_command("on_child_form_ok")
                    return
                if target == "message":
                    dlg = wx.MessageDialog(
                        self, s, is_null(mp.title, title), style=wx.OK
                    )
                    dlg.ShowModal()
                    return

                if target == "code":
                    script = mp.get_body()[1]
                    app = wx.GetApp()
                    main_window = app.GetTopWindow()
                    desktop = (main_window.desktop,)
                    mgr = main_window._mgr
                    menu_bar = main_window.get_menu_bar()
                    tool_bars = main_window.get_tool_bar()
                    exec(script.elem.text.replace("\r", ""))
                    return

    def create_websocket(self, url, callback):
        app = wx.GetApp()
        self.websockets.append((url, callback))
        app.create_websocket(url, callback)

    def websocket_send(self, url, data):
        app = wx.GetApp()

        async def _send():
            nonlocal app, url, data
            await app.websocket_send(url, data)

        app.StartCoroutine(_send, app.GetTopWindow())

    def signal_from_child(self, child_object, signal):
        pass

    def get_evt_ind(self):
        if self.evt_ind == -1:
            self.evt_ind = wx.ID_HIGHEST
        else:
            self.evt_ind += 1
        return self.evt_ind

    def set_acc_key_tab(self, win, tab):
        """Set accelerator table for window

        Args:
            win - window which handle accelerator keys
            tab - list or tuple of accelerator elements. Each emement has structure: (flag, keycode, callback fun)
        """
        if win in self.acc_tabs:
            self.acc_tabs[win].append(tab)
        else:
            self.acc_tabs[win] = [
                tab,
            ]

    def _build_acc_tab(self, to_win=None):
        ret = []
        if to_win:
            obj = to_win
        else:
            obj = self
        for win in self.acc_tabs.keys():
            tab = []
            for pos in self.acc_tabs[win]:
                tab += pos
            ret.append((win, tab))
        obj._set_acc_key_tab(ret)

    def _set_acc_key_tab(self, tabs):
        tab2 = []

        for tab_pos in tabs:
            win = tab_pos[0]
            tab = tab_pos[1]

            id_start = self.get_evt_ind()
            for pos in tab:
                tab2.append((pos[0], pos[1], self.get_evt_ind()))

            if wx.Platform == "__WXMSW__":
                win.SetAcceleratorTable(wx.AcceleratorTable(tab2))
                tab2 = []

                def make_fun(tab, id_start):
                    def on_command(event):
                        id = event.GetId()
                        ind = id - id_start - 1
                        if ind >= 0 and ind < len(tab):
                            cmd = tab[ind][2]
                            cmd(event)
                        else:
                            event.Skip()

                    return on_command

                win.Bind(wx.EVT_MENU, make_fun(tab, id_start))
            else:
                if not self.acc_tab:
                    self.acc_tab = []
                for pos in tab:
                    self.acc_tab.append(
                        list(pos)
                        + [
                            win,
                        ]
                    )

                if not win.acc_tab_bind:
                    win.Bind(wx.EVT_KEY_DOWN, self.on_acc_key_down)
                    win.acc_tab_bind = True

        if wx.Platform != "__WXMSW__":
            self.SetAcceleratorTable(wx.AcceleratorTable(tab2))

    def on_acc_key_down(self, event):
        if event.KeyCode == 307:
            return
        for a in self.acc_tab:
            if event.KeyCode == a[1]:
                if (not event.AltDown()) and (a[0] & wx.ACCEL_ALT):
                    continue
                if (not event.ControlDown()) and (a[0] & wx.ACCEL_CTRL):
                    continue
                if (not event.ShiftDown()) and (a[0] & wx.ACCEL_SHIFT):
                    continue

                if event.AltDown() and not (a[0] & wx.ACCEL_ALT):
                    continue
                if event.ControlDown() and not (a[0] & wx.ACCEL_CTRL):
                    continue
                if event.ShiftDown() and not (a[0] & wx.ACCEL_SHIFT):
                    continue
                p = wx.Window.FindFocus()
                while p:
                    if p == a[3]:
                        a[2](event)
                        return
                    p = p.GetParent()
                return
        event.Skip()



import wx

"""Module contains all user defined events
"""


ID_START = wx.NewId()

ID_NEXTTAB = wx.NewId()
ID_PREVTAB = wx.NewId()
ID_CLOSETAB = wx.NewId()
ID_NEXTPAGE = wx.NewId()
ID_PREVPAGE = wx.NewId()
ID_REFRESHTAB = wx.NewId()
ID_EXIT = wx.NewId()
ID_RESET = wx.NewId()
ID_UNDO = wx.NewId()
ID_REDO = wx.NewId()
ID_CUT = wx.NewId()
ID_COPY = wx.NewId()
ID_PASTE = wx.NewId()
ID_SELALL = wx.NewId()
ID_HELP = wx.NewId()
ID_GOTOHEADER = wx.NewId()
ID_GOTOPANEL = wx.NewId()
ID_GOTOFOOTER = wx.NewId()
ID_GOTODESKTOP = wx.NewId()
ID_SHOWHEADER = wx.NewId()
ID_SHOWPANEL = wx.NewId()
ID_SHOWFOOTER = wx.NewId()
ID_SHOWDESKTOP = wx.NewId()
ID_SHOWTOOLBAR1 = wx.NewId()
ID_SHOWTOOLBAR2 = wx.NewId()
ID_SHOWSTATUSBAR = wx.NewId()
ID_RETVALUE = wx.NewId()

ID_ZOOM = wx.NewId()
ID_PAGE_SOURCE = wx.NewId()
ID_PAGE_BLOCK = wx.NewId()
ID_PAGE_VIEW = wx.NewId()
ID_HISTORY_SHOW = wx.NewId()
ID_CLEAR_HISTORY = wx.NewId()
ID_BOOKMARK_ADD = wx.NewId()
ID_SHOW_BOOKMARKS = wx.NewId()
ID_SHOW_DOWNLOAD = wx.NewId()
ID_DOWNLOAD_OPTIONS = wx.NewId()
ID_GENERAL_OPTIONS = wx.NewId()
ID_DOWNLOAD_OPTIONS = wx.NewId()
ID_FIND = wx.NewId()
ID_REPLACE = wx.NewId()
ID_PRINT = wx.NewId()
ID_PRINT_PREVIEW = wx.NewId()
ID_PRINT_SETTINGS = wx.NewId()
ID_WEB_BACK = wx.NewId()
ID_WEB_FORWARD = wx.NewId()
ID_WEB_STOP = wx.NewId()
ID_WEB_REFRESH = wx.NewId()
ID_WEB_ADDBOOKMARK = wx.NewId()
ID_WEB_NEW_WINDOW = wx.NewId()
ID_WEB_SOURCE = wx.NewId()
ID_WEB_UP = wx.NewId()
ID_WEB_DOWN = wx.NewId()
ID_WEB_EDIT = wx.NewId()
ID_TASKBAR_SHOW = wx.NewId()
ID_TASKBAR_HIDE = wx.NewId()
ID_TASKBAR_CLOSE = wx.NewId()
ID_TASKBAR_TOGGLE = wx.NewId()
ID_LOAD = wx.NewId()
ID_SAVE = wx.NewId()
ID_SAVE_AS = wx.NewId()

ID_END = wx.NewId()



import sys

import wx

from pytigon_lib.schtools import schjson

_ = wx.GetTranslation


CMD_INFO = 1
CMD_PAGE = 2
CMD_COUNT = 3
CMD_SYNC = 4
CMD_AUTO = 5
CMD_RECASSTR = 6
CMD_EXEC = 7


def process_post_parm(obj):
    ret = {}
    for key, value in list(obj.items()):
        ret[key] = schjson.dumps(value)
    return ret


class DataProxy:
    def __init__(self, http, tabaddress):
        self.max_count = 1000000
        self.var_count = -1
        self.tabaddress = tabaddress
        self.tabaddress0 = self.tabaddress
        self.parm = ""
        self.parent = None
        self.http = http
        self.col_types2 = []
        self.parm = dict()
        self.is_valid = True

        response = self.http.post(
            self.parent,
            self.tabaddress,
            process_post_parm(
                {
                    "cmd": CMD_INFO,
                }
            ),
        )
        ret = schjson.loads(response.str())

        self.col_names = ret["col_names"]
        self.col_types = ret["col_types"]
        self.default_rec = ret["default_rec"]
        self.col_size = ret["col_length"]

        for i in range(0, len(self.col_size)):
            if (self.col_size)[i] > 32:
                (self.col_size)[i] = 32

        self.auto_cols = ret["auto_cols"]

        for col in self.col_types:
            self.col_types2.append(col.split(":")[0])

        self.tab_conw = {
            "long": self.conw_long,
            "string": self.conw_none,
            "double": self.conw_float,
            "bool": self.conw_bool,
            "choice": self.conw_none,
            "x": self.conw_none,
        }

    def set_parent(self, parent):
        self.parent = parent

    def set_address(self, address):
        self.tabaddress = self.tabaddress0 + address

    def get_address(self):
        return self.tabaddress

    def set_address_parm(self, parm):
        if self.parm == parm:
            return False
        else:
            self.parm = parm
            return True

    def get_address_parm(self):
        return self.parm

    def conw_long(self, l):
        if l:
            return int(l)
        else:
            return None

    def conw_none(self, n):
        if n != None:
            return n
        else:
            return ""

    def conw_float(self, f):
        if f:
            return float(f)
        else:
            return None

    def conw_bool(self, b):
        if b:
            return bool(b)
        else:
            return None

    def _reformat_rec(self, rec):
        ret = []
        i = 0
        for col in rec:
            ret.append((self.tab_conw)[(self.col_types2)[i]](col))
            i = i + 1
        return ret

    def set_parm(self, key, value):
        (self.parm)[key] = value

    def get_page(self, nrPage):
        c = {"cmd": CMD_PAGE, "nr": nrPage}
        if self.parm:
            for key, value in list(self.parm.items()):
                c[key] = value
        response = self.http.post(self.parent, self.tabaddress, process_post_parm(c))

        try:
            page = schjson.loads(response.str())
            retpage = page["page"]
        except:
            retpage = None
            # from pytigon_lib.schhttptools.httperror import http_error
            # http_error(wx.GetApp().GetTopWindow(), self.http.str())
            self.http.show(wx.GetApp().GetTopWindow())
            retpage = []
        return retpage

    def get_max_count(self):
        return self.max_count

    def get_count(self):
        parm = {"cmd": CMD_COUNT}
        if "value" in self.parm:
            parm["value"] = self.parm["value"]
        response = self.http.post(self.parent, self.tabaddress, process_post_parm(parm))
        s = response.str()
        ret = schjson.loads(s)
        self.max_count = int(ret["count"])
        return self.max_count

    def sync_data(self, listaRecUpdate, listaRecInsert, listaRecDelete):
        update = schjson.dumps(listaRecUpdate)
        insert = schjson.dumps(listaRecInsert)
        delete = schjson.dumps(listaRecDelete)
        c = {"cmd": CMD_SYNC, "update": update, "insert": insert, "delete": delete}
        self.http.post(self.parent, self.tabaddress, process_post_parm(c))

    def auto_update(self, col_name, col_names, rec):
        """Return transformed row after current row is changed"""

        col_name2 = schjson.dumps(col_name)
        col_names2 = schjson.dumps(col_names)
        rec2 = schjson.dumps(rec)

        c = {
            "cmd": CMD_AUTO,
            "col_name": col_name2,
            "col_names": col_names2,
            "rec": rec2,
        }

        response = self.http.post(self.parent, self.tabaddress, process_post_parm(c))
        ret = schjson.loads(response.str())
        if ret == None:
            return rec
        else:
            return ret["rec"]

    def clone(self):
        c = DataProxy(self.http, self.tabaddress)
        c.set_address_parm(self.get_address_parm())
        return c

    def exec(self, parm):
        c = {"cmd": CMD_EXEC, "value": parm}
        response = self.http.post(self.parent, self.tabaddress, process_post_parm(c))
        ret = schjson.loads(response.str())
        return ret

    def get_default_rec(self):
        return self.default_rec

    def GetRecAsStr(self, nrRec):
        response = self.http.post(
            self.parent,
            self.tabaddress,
            process_post_parm({"cmd": CMD_RECASSTR, "nr": nrRec}),
        )
        ret = schjson.loads(response.str())
        return ret["recasstr"]

    def GetColNames(self):
        return self.col_names

    def GetAutoCols(self):
        return self.auto_cols

    def GetColTypes(self):
        return self.col_types

    def GetColSize(self):
        return self.col_size

    def GetColIcons(self):
        return None



import wx
from pytigon_gui.guilib.events import *
from autocomplete import TextCtrlAutoComplete
from pytigon_gui.guilib.image import bitmaps_from_art_id
from pytigon_gui.guilib.tools import colour_to_html, find_focus_in_form
from wx.lib.agw import flatmenu as FM

_ = wx.GetTranslation

if wx.Platform == "__WXMSW__":
    MswStyle = True
else:
    MswStyle = True

tab = [
    "www.allegro.pl",
    "www.onet.pl",
    "www.google.pl",
    "www.sourceforge.org",
    "www.freshmeat.net",
    "127.0.0.2",
    "127.0.0.1:8000",
    "books.evc-cit.info/index.html",
    "www.un.org/Depts/Cartographic/map/profile/poland.pdf",
]

TYPE_TOOLBAR = 0
TYPE_BUTTONBAR = 1
TYPE_PANELBAR = 2


class WebHistory(object):
    def __iter__(self):
        for x in tab:
            yield x

    def __getitem__(self, id):
        if id < 4:
            return tab[id]
        else:
            x = id / 0
            return None

    def __len__(self):
        return 4

    def __contains__(self, x):
        if x in tab:
            return True
        else:
            return False


class SchAutoComplete(TextCtrlAutoComplete):
    def __init__(self, *argi, **argv):
        self.dynamic_choices = WebHistory()
        argv["choices"] = self.dynamic_choices
        TextCtrlAutoComplete.__init__(self, *argi, **argv)
        self.SetEntryCallback(self.setDynamicChoices)
        self.SetMatchFunction(self.match)
        self.Bind(wx.EVT_KEY_DOWN, self.on_return_key_down, self)

    def on_return_key_down(self, event):
        if not self.dropdown.IsShown() and event.GetKeyCode() == wx.WXK_RETURN:
            self.fun(self.GetValue())
        else:
            event.Skip()

    def set_callback_fun(self, fun):
        self.fun = fun

    def match(self, text, choice):
        t = text.lower()
        c = choice.lower()
        if c.startswith(t):
            return True
        if c.startswith(r"http://"):
            c = c[7:]
        if c.startswith(t):
            return True
        if c.startswith("www."):
            c = c[4:]
        return c.startswith(t)

    def set_dynamic_choices(self):
        ctrl = self
        text = ctrl.GetValue().lower()
        current_choices = ctrl.GetChoices()
        choices = [
            choice for choice in self.dynamic_choices if self.match(text, choice)
        ]
        if choices != current_choices:
            ctrl.SetChoices(choices)

    def _set_value_from_selected(self):
        x = TextCtrlAutoComplete._setValueFromSelected(self)
        self.fun(self.GetValue())
        return x


class EmpytControl(object):
    def __init__(self):
        self.value = None

    def SetValue(self, value):
        self.value = value

    def GetValue(self):
        return self.value


class EmptyBar(object):
    def __init__(self):
        pass

    def refr(self):
        pass


class StandardButtons(object):
    def __init__(self, toolbar_interface, gui_style):
        self.toolbar_interface = toolbar_interface
        self.gui_style = gui_style
        self.ti = self.toolbar_interface
        self.tbs = self.ti.toolbars
        self._object_list = {}
        self.webobject = None
        self.file = True if "file" in self.gui_style else False
        self.clipboard = True if "clipboard" in self.gui_style else False
        self.operations = True if "operations" in self.gui_style else False
        self.browse = True if "browse" in self.gui_style else False
        self.nav = True if "nav" in self.gui_style else False
        self.address_bar = True if "address_bar" in self.gui_style else False
        self.bg = colour_to_html(wx.SystemSettings.GetColour(wx.SYS_COLOUR_3DFACE))
        self.bg_info = colour_to_html(wx.SystemSettings.GetColour(wx.SYS_COLOUR_INFOBK))
        self.bg_h = colour_to_html(wx.SystemSettings.GetColour(wx.SYS_COLOUR_HIGHLIGHT))
        self.progress = 0
        self.info = ""
        self.address_txt = ""

    def make_ui_handler(self, fun_name, event_id, propagete=False):
        def on_update_ui(event):
            # win = wx.Window.FindFocus()
            win = find_focus_in_form()
            while win:
                if hasattr(win, fun_name):
                    if getattr(win, fun_name)():
                        event.Enable(True)
                        return
                if propagete:
                    win = win.GetParent()
                else:
                    break
            event.Enable(False)

        self.toolbar_interface.parent.Bind(wx.EVT_UPDATE_UI, on_update_ui, id=event_id)

    def make_handler(self, fun_name, event_id, propagate=False):
        def on_command(event):
            # win = wx.Window.FindFocus()
            win = find_focus_in_form()
            while win:
                if hasattr(win, fun_name):
                    getattr(win, fun_name)()
                if propagate:
                    win = win.GetParent()
                else:
                    break

        self.toolbar_interface.bind(on_command, id=event_id)

    def make_handlers(self, event_id, fun_name, fun_name2=None, propagate=False):
        self.make_handler(fun_name, event_id, propagate)
        if fun_name2:
            self.make_ui_handler(fun_name2, event_id, propagate)
        else:

            def on_update_ui(event):
                event.Enable(True)

            self.toolbar_interface.parent.Bind(
                wx.EVT_UPDATE_UI, on_update_ui, id=event_id
            )

    def create_file_panel(self, toolbar_page):
        if self.file:
            bar = self.ti.create_panel_in_main_page(_("File"), TYPE_TOOLBAR)
            self.tbs["file"] = {}
            self.tbs["file"]["bar"] = bar
            if "exit" in self.gui_style:
                self.tbs["file"]["exit"] = bar.add_simple_tool(
                    wx.ID_EXIT,
                    _("Exit"),
                    bitmaps_from_art_id(wx.ART_QUIT, wx.Size(32, 32)),
                )
            test = False
            if "open" in self.gui_style:
                bar.add_separator()
                self.tbs["file"]["open"] = bar.add_simple_tool(
                    wx.ID_OPEN,
                    _("Open"),
                    bitmaps_from_art_id(wx.ART_FILE_OPEN, wx.Size(32, 32)),
                )
                test = True
            if "save" in self.gui_style:
                if not test:
                    bar.add_separator()
                self.tbs["file"]["save"] = bar.add_simple_tool(
                    wx.ID_SAVE,
                    _("Save"),
                    bitmaps_from_art_id(wx.ART_FILE_SAVE, wx.Size(32, 32)),
                )
                self.make_handlers(wx.ID_SAVE, "Save", "CanSave", propagate=True)
                if "save_as" in self.gui_style:
                    self.tbs["file"]["save_as"] = bar.add_simple_tool(
                        wx.ID_SAVEAS,
                        _("Save as"),
                        bitmaps_from_art_id(wx.ART_FILE_SAVE_AS, wx.Size(32, 32)),
                    )
                    self.make_handlers(
                        wx.ID_SAVEAS, "SaveAs", "CanSaveAs", propagate=True
                    )
                test = True
            if "print" in self.gui_style:
                if test:
                    bar.add_separator()
                    self.tbs["file"]["print"] = bar.AddHybridTool(
                        ID_PRINT,
                        _("Print"),
                        bitmaps_from_art_id(wx.ART_PRINT, wx.Size(32, 32)),
                    )
                    bar.add_separator()
                else:
                    self.tbs["file"]["print"] = bar.AddHybridTool(
                        ID_PRINT,
                        _("Print"),
                        bitmaps_from_art_id(wx.ART_PRINT, wx.Size(32, 32)),
                    )
                self.make_handlers(ID_PRINT, "Print", "CanPrint", propagate=True)
                self.ti.bind_dropdown(self.on_print, ID_PRINT)

    def create_edit_panel(self, toolbar_page):
        if self.clipboard:
            self.tbs["clipboard"] = {}
            bar = self.ti.create_panel_in_main_page(_("Clipboard"), TYPE_BUTTONBAR)
            bar.Refr = self.clipboard_refr
            self.tbs["clipboard"]["bar"] = bar
            self.tbs["clipboard"]["copy"] = bar.add_simple_tool(
                wx.ID_COPY, _("Copy"), bitmaps_from_art_id(wx.ART_COPY, wx.Size(32, 32))
            )
            self.tbs["clipboard"]["cut"] = bar.add_simple_tool(
                wx.ID_CUT, _("Cut"), bitmaps_from_art_id(wx.ART_CUT, wx.Size(32, 32))
            )
            self.tbs["clipboard"]["paste"] = bar.add_simple_tool(
                wx.ID_PASTE,
                _("Paste"),
                bitmaps_from_art_id(wx.ART_PASTE, wx.Size(32, 32)),
            )

            self.make_handlers(wx.ID_COPY, "Copy", "CanCopy")
            self.make_handlers(wx.ID_CUT, "Cut", "CanCut")
            self.make_handlers(wx.ID_PASTE, "Paste", "CanPaste")

    def create_operations_panel(self, toolbar_page):
        if self.operations:
            self.tbs["operations"] = {}
            bar = self.ti.CreatePanelInMainPage(_("Operations"), TYPE_TOOLBAR)
            self.tbs["operations"]["bar"] = bar
            self.tbs["operations"]["undo"] = bar.add_simple_tool(
                wx.ID_UNDO, _("Undo"), bitmaps_from_art_id(wx.ART_UNDO, wx.Size(32, 32))
            )
            self.tbs["operations"]["redo"] = bar.add_simple_tool(
                wx.ID_REDO, _("Redo"), bitmaps_from_art_id(wx.ART_REDO, wx.Size(32, 32))
            )
            self.tbs["operations"]["find"] = bar.add_hybrid_tool(
                ID_FIND, _("Find"), bitmaps_from_art_id(wx.ART_FIND, wx.Size(32, 32))
            )
            bar.AddSeparator()
            self.tbs["operations"]["filemanager"] = bar.add_simple_tool(
                wx.ID_ANY,
                _("File manager"),
                bitmaps_from_art_id(wx.ART_FOLDER_OPEN, wx.Size(32, 32)),
            )
            self.tbs["operations"]["editor"] = bar.add_simple_tool(
                wx.ID_ANY,
                _("Editor"),
                bitmaps_from_art_id(wx.ART_NORMAL_FILE, wx.Size(32, 32)),
            )
            self.ti.bind_dropdown(self.on_find, ID_FIND)

            self.make_handlers(wx.ID_UNDO, "Undo", "CanUndo")
            self.make_handlers(wx.ID_REDO, "Redo", "CanRedo")
            self.make_handlers(ID_FIND, "Find", "CanFind")

    def create_browse_panel(self, toolbar_page):
        if self.browse or self.nav:
            self.tbs["browser"] = {}
            bar = self.ti.create_panel_in_main_page(_("Browser"), TYPE_BUTTONBAR)
            bar.Refr = self.web_refr
            self.webobject = None
            self.tbs["browser"]["bar"] = bar
            self.tbs["browser"]["back"] = bar.add_simple_tool(
                ID_WEB_BACK,
                _("Back"),
                bitmaps_from_art_id(wx.ART_GO_BACK, wx.Size(32, 32)),
            )
            self.tbs["browser"]["forward"] = bar.add_simple_tool(
                ID_WEB_FORWARD,
                _("Forward"),
                bitmaps_from_art_id(wx.ART_GO_FORWARD, wx.Size(32, 32)),
            )
            self.make_handlers(ID_WEB_BACK, "WebBack", "CanWebBack", propagate=True)
            self.make_handlers(
                ID_WEB_FORWARD, "WebForward", "CanWebForward", propagate=True
            )

            if self.browse:
                self.tbs["browser"]["stop"] = bar.add_simple_tool(
                    ID_WEB_STOP,
                    _("Stop"),
                    bitmaps_from_art_id(wx.ART_CROSS_MARK, wx.Size(32, 32)),
                )
                self.tbs["browser"]["refresh"] = bar.add_simple_tool(
                    ID_WEB_REFRESH,
                    _("Refresh"),
                    bitmaps_from_art_id(wx.ART_GO_TO_PARENT, wx.Size(32, 32)),
                )
                self.tbs["browser"]["addbookmark"] = bar.add_simple_tool(
                    ID_WEB_ADDBOOKMARK,
                    _("Add bookmark"),
                    bitmaps_from_art_id(wx.ART_ADD_BOOKMARK, wx.Size(32, 32)),
                )
                self.tbs["browser"]["newpage"] = bar.add_simple_tool(
                    ID_WEB_NEW_WINDOW,
                    _("New page"),
                    bitmaps_from_art_id(wx.ART_NEW, wx.Size(32, 32)),
                )

                self.make_handlers(ID_WEB_STOP, "WebStop", "CanWebStop", propagate=True)
                self.make_handlers(
                    ID_WEB_REFRESH, "WebRefresh", "CanWebRefresh", propagate=True
                )
                self.make_handlers(
                    ID_WEB_ADDBOOKMARK,
                    "WebAddBookmark",
                    "CanWebAddBookmark",
                    propagate=True,
                )
                self.make_handlers(ID_WEB_NEW_WINDOW, "WebNewWindow", propagate=None)
        else:
            self.tbs["browser"] = {}
            self.tbs["browser"]["bar"] = EmptyBar()

    def create_address_panel(self, toolbar_page):
        if self.address_bar:
            bar = self.ti.create_panel_in_main_page(_("Address"), TYPE_PANELBAR)
            if hasattr(bar, "panel"):
                width = 350
                panel = bar.panel
                self.address = SchAutoComplete(panel, size=wx.Size(width, -1))
                self.address.SetCallbackFun(self.OpenPage)
                self.tbs["browser"]["address"] = self.address
                self.webmenu = FM.FlatMenuBar(panel, spacer=5)
                self.webmenu.GetRendererManager().SetTheme(FM.Style2007)
                self.tbs["browser"]["webmenu"] = self.webmenu
                webmenu_source = self.WebMenu()
                self.web_info = wx.html.HtmlWindow(
                    panel, style=wx.html.HW_SCROLLBAR_NEVER, size=(width, 25)
                )
                self.web_info.SetBorders(0)
                self._info_clear()
                self.MemkeMenu(self.webmenu, webmenu_source)
                controls = (self.address, self.web_info, self.webmenu)
                controls = (self.webmenu, self.web_info, self.address)
                bar.AddPanel(controls)
            else:
                self.tbs["browser"]["address"] = EmpytControl()
                self.web_info = None
        else:
            self.tbs["browser"]["address"] = EmpytControl()
            self.web_info = None

    def load_page(self, address):
        if "browser" in self._object_list and self._object_list["browser"]:
            self._object_list["browser"].Go(address)

    def _title_from_address(self, address):
        x = address.replace(r"http://", "")
        if len(x) > 20:
            x = x[:16] + " ..."
        return x

    def _status(self, tab):
        start_status = []
        dis = self.ti.status_tool_disabled()
        for pos in tab:
            start_status.append(self.tbs[pos[1]][pos[2]].state & dis)
            if pos[0]:
                self.tbs[pos[1]][pos[2]].state &= ~dis
            else:
                self.tbs[pos[1]][pos[2]].state |= dis
        i = 0
        for pos in tab:
            if start_status[i] != self.tbs[pos[1]][pos[2]].state & dis:
                self.tbs[pos[1]]["bar"].Refresh()
                return
            i += 1

    def clipboard_refr(self, editctrl):
        if self.clipboard:
            tab_status = []
            if editctrl:
                tab_status.append(
                    (
                        hasattr(editctrl, "CanCopy") and editctrl.CanCopy(),
                        "clipboard",
                        "copy",
                    )
                )
                tab_status.append(
                    (
                        hasattr(editctrl, "CanCut") and editctrl.CanCut(),
                        "clipboard",
                        "cut",
                    )
                )
                tab_status.append(
                    (
                        hasattr(editctrl, "CanPaste") and editctrl.CanCut(),
                        "clipboard",
                        "paste",
                    )
                )
            else:
                tab_status.append((False, "clipboard", "copy"))
                tab_status.append((False, "clipboard", "cut"))
                tab_status.append((False, "clipboard", "paste"))
            return self._status(tab_status)

    def opeartions_refr(self, editctrl):
        if self.operations:
            tab_status = []
            if editctrl:
                tab_status.append(
                    (
                        hasattr(editctrl, "CanRedo") and editctrl.CanCopy(),
                        "operations",
                        "redo",
                    )
                )
                tab_status.append(
                    (
                        hasattr(editctrl, "CanUndo") and editctrl.CanCut(),
                        "operations",
                        "undo",
                    )
                )
            else:
                tab_status.append((False, "operations", "redo"))
                tab_status.append((False, "operations", "undo"))
            return self._status(tab_status)

    def web_refr(self, webobject):
        if self.browse or self.nav:
            tab_status = []
            if webobject:
                tab_status.append((webobject.CanGoBack(), "browser", "back"))
                tab_status.append((webobject.CanGoForward(), "browser", "forward"))
                if self.browse:
                    tab_status.append((True, "browser", "stop"))
                    tab_status.append((True, "browser", "refresh"))
                    tab_status.append((True, "browser", "addbookmark"))
                    tab_status.append((True, "browser", "newpage"))
            else:
                tab_status.append((False, "browser", "back"))
                tab_status.append((False, "browser", "forward"))
                if self.browse:
                    tab_status.append((False, "browser", "stop"))
                    tab_status.append((False, "browser", "refresh"))
                    tab_status.append((False, "browser", "addbookmark"))
                    tab_status.append((False, "browser", "newpage"))
            return self._status(tab_status)

    def address_refr(self, panel):
        if self.address_bar:
            if panel is not None and hasattr(panel.body, "WEB"):
                ctrl = panel.body.WEB
                if hasattr(ctrl, "GetStatus"):
                    status = ctrl.GetStatus()
                    if status["progress"] >= 0:
                        self._info_proc(
                            status["txt"].replace("INFO:", "").replace("ERROR:", ""),
                            status["progress"],
                        )
                    else:
                        if status["txt"] and status["txt"] != "":
                            if "ERROR:" in status["txt"]:
                                self._info_alarm(status["txt"].replace("ERROR:", ""))
                            else:
                                self._info_txt(status["txt"].replace("INFO:", ""))
                        else:
                            self._info_clear()
                    if self.address_txt != status["address"]:
                        if status["address"]:
                            self.address.SetValue(status["address"])
                            self.address_txt = status["address"]
                        else:
                            self.address.SetValue("")
                            self.address_txt = ""
            else:
                if self.address_txt != "":
                    self.address.SetValue("")
                    self.address_txt = ""
                self._info_clear()

    def _info_proc(self, txt, proc):
        if self.web_info and (txt != self.info or proc != self.progress):
            if txt and txt != "":
                self.web_info.SetPage(
                    "<body bgcolor='%s'><table width='100%%'><tr><td bgcolor='%s' width='%s%%'></td><td ></td></tr><tr><td colspan='2' td bgcolor='%s'><small>%s</small></td></tr></table></body>"
                    % (self.bg, self.bg_h, proc, self.bg_info, txt)
                )
            else:
                self.web_info.SetPage(
                    "<body bgcolor='%s'><table width='100%%'><tr><td bgcolor='%s' width='%s%%'></td><td ></td></tr></table></body>"
                    % (self.bg, self.bg_h, proc)
                )
            self.info = txt
            self.progress = proc

    def _info_txt(self, txt):
        if self.web_info and txt != self.info:
            self.web_info.SetPage(
                "<body bgcolor='%s'><table width='100%%'><tr><td bgcolor='%s'><strong><small>%s</small></strong></tr></td></table></body>"
                % (self.bg, self.bg_info, txt)
            )
            self.info = txt
            self.progress = -1

    def _info_alarm(self, txt):
        if self.web_info and txt != self.info:
            self.web_info.SetPage(
                "<body bgcolor='%s'><table width='100%%'><tr><td bgcolor='%s'><strong><small>%s</small></strong></tr></td></table></body>"
                % (self.bg, self.bg_info, txt)
            )
            self.info = txt
            self.progress = -1

    def _info_clear(self):
        if self.web_info and (self.info != "" or self.progress != -1):
            self.web_info.SetPage("<body bgcolor='%s'></body>" % self.bg)
            self.info = ""
            self.progress = -1

    def open_page(self, page):
        name = self._title_from_address(page)
        okno = (
            wx.GetApp()
            .GetTopWindow()
            .new_main_page("^standard/webview/widget_web.html", name)
        )
        okno.body.WEB.Go(page)
        return True

    def on_print(self, event):
        menu = FM.FlatMenu()
        menu.Append(wx.ID_ANY, _("Print"), _("Print"), wx.ITEM_NORMAL)
        menu.Append(wx.ID_ANY, _("Print preview"), _("Print preview"), wx.ITEM_NORMAL)
        menu.Append(
            wx.ID_ANY, _("Printer settings"), _("Printer settings"), wx.ITEM_NORMAL
        )
        self.ti.popup_menu(event, menu)

    def on_find(self, event):
        menu = FM.FlatMenu()
        menu.Append(wx.ID_ANY, _("Find"), _("Find"), wx.ITEM_NORMAL)
        menu.Append(wx.ID_ANY, _("Replace"), _("Replace"), wx.ITEM_NORMAL)
        event.PopupMenu(menu)

    def memke_menu(self, menubar, menulist):
        for item in menulist:
            menu = FM.FlatMenu()
            for pos in item[1]:
                menuitem = FM.FlatMenuItem(menu, pos[1], pos[0], pos[0], wx.ITEM_NORMAL)
                menu.AppendItem(menuitem)
            menubar.Append(menu, item[0])

    def web_menu(self):
        menu = (  # ('Page source', ID_PAGE_SOURCE),
            (
                _("View"),
                (
                    (_("Zoom"), ID_ZOOM),
                    (_("Block elements"), ID_PAGE_BLOCK),
                    (_("View elements"), ID_PAGE_VIEW),
                    (_("View page source"), ID_WEB_SOURCE),
                    (_("Edit page"), ID_WEB_EDIT),
                ),
            ),
            (
                _("History"),
                (
                    (_("Show history panel"), ID_HISTORY_SHOW),
                    (_("Clear history"), ID_CLEAR_HISTORY),
                ),
            ),
            (
                _("Bookmarks"),
                (
                    (_("Add to bookmarks"), ID_BOOKMARK_ADD),
                    (_("Show bookmarks panel"), ID_SHOW_BOOKMARKS),
                ),
            ),
            (
                _("Downloads"),
                (
                    (_("Show download panel"), ID_SHOW_DOWNLOAD),
                    (_("Options"), ID_DOWNLOAD_OPTIONS),
                ),
            ),
            (
                _("Tools"),
                (
                    (_("Show download panel"), ID_SHOW_DOWNLOAD),
                    x(_("Options"), ID_DOWNLOAD_OPTIONS),
                ),
            ),
            (
                _("Options"),
                (
                    (_("General"), ID_GENERAL_OPTIONS),
                    (_("Black list"), ID_DOWNLOAD_OPTIONS),
                ),
            ),
        )
        return menu

    def realize(self):
        pass



import wx

_ = wx.GetTranslation


class LoginDialog(wx.Dialog):
    """Helper class to create login dialog"""

    def __init__(
        self,
        parent,
        id,
        title,
        size=wx.DefaultSize,
        pos=wx.DefaultPosition,
        style=wx.DEFAULT_DIALOG_STYLE,
        use_metal=False,
        username=None,
    ):
        wx.Dialog.__init__(self, parent, id, title, pos, size, style)
        self.CenterOnParent()

        if "wxMac" in wx.PlatformInfo and use_metal:
            self.SetExtraStyle(wx.DIALOG_EX_METAL)
        sizer = wx.GridBagSizer(5, 2)
        self.message = wx.StaticText(self, -1, "")
        font = wx.Font(
            10, wx.FONTFAMILY_DEFAULT, wx.FONTSTYLE_NORMAL, wx.FONTWEIGHT_BOLD
        )
        self.message.SetFont(font)
        label1 = wx.StaticText(self, -1, _("User name:"))
        self.text1 = wx.TextCtrl(self, -1, "", size=(180, -1))
        if username:
            self.text1.SetValue(username)
        label2 = wx.StaticText(self, -1, _("Password:"))
        self.text2 = wx.TextCtrl(self, -1, "", size=(180, -1), style=wx.TE_PASSWORD)
        btnsizer = wx.StdDialogButtonSizer()
        btn = wx.Button(self, wx.ID_OK)
        btn.SetDefault()
        btnsizer.AddButton(btn)
        btn = wx.Button(self, wx.ID_CANCEL)
        btnsizer.AddButton(btn)
        btnsizer.Realize()
        sizer.Add(self.message, (0, 0))
        sizer.Add(5, 5, (1, 0))
        sizer.Add(label1, (1, 1))
        sizer.Add(self.text1, (1, 2))
        sizer.Add(label2, (2, 1))
        sizer.Add(self.text2, (2, 2))
        sizer.Add(btnsizer, (3, 2), (1, 1))
        sizer.Add(5, 5, (4, 4))
        self.SetSizer(sizer)
        sizer.Fit(self)
        if username:
            self.text2.SetFocus()



"""
Module contain helper classes for SELEC2 widget.
"""

import string
import wx

from pytigon_lib.schtools import schjson
from pytigon_gui.guictrl.basectrl import SchBaseCtrl

_ = wx.GetTranslation


class ListBoxNoFocus(wx.ListBox):
    def CanAcceptFocus(self):
        return False


class Select2Popup(wx.MiniFrame):
    def __init__(
        self,
        parent,
        id,
        title,
        pos,
        size,
        style,
        combo,
        field_id,
        url=None,
        minimum_input_length=2,
    ):
        from pytigon_gui.guiframe.page import SchPage

        self.combo = combo
        self.point = pos
        self.field_id = field_id
        self.url = url
        if minimum_input_length:
            self.minimum_input_length = minimum_input_length
        else:
            minimum_input_length = 0

        wx.MiniFrame.__init__(self, parent, id, title, pos, size, wx.RESIZE_BORDER)

        self.edit_ctrl = wx.TextCtrl(
            self, size=(440, -1), style=wx.TE_PROCESS_ENTER | wx.TE_PROCESS_TAB
        )
        self.list_ctrl = ListBoxNoFocus(self, size=(440, 200), style=wx.LB_SINGLE)

        self.edit_ctrl.Bind(wx.EVT_KEY_DOWN, self.on_key_down)
        self.edit_ctrl.Bind(wx.EVT_TEXT_ENTER, self.on_enter)
        self.list_ctrl.Bind(wx.EVT_LISTBOX_DCLICK, self.on_enter)
        self.list_ctrl.Bind(wx.EVT_LISTBOX, self.on_enter)

        self.Bind(wx.EVT_ACTIVATE, self.on_activate)
        self.Bind(wx.EVT_TEXT, self.on_text)

        box = wx.BoxSizer(wx.VERTICAL)
        box.Add(self.edit_ctrl)
        box.Add(self.list_ctrl, 1, wx.ALL | wx.GROW, 1)
        self.SetSizer(box)
        self.SetAutoLayout(True)
        self.Fit()

    def on_enter(self, event):
        id = self.list_ctrl.GetSelection()
        if id != wx.NOT_FOUND:
            self.Dismiss()
            item_id = self.list_ctrl.GetClientData(id)
            item_str = self.list_ctrl.GetString(id)
            self.combo.set_value(item_id, item_str)

    def on_activate(self, event):
        if not event.GetActive():
            self.Hide()
            self.combo.SetFocus()
        event.Skip()

    def on_key_down(self, event):
        if event.KeyCode == wx.WXK_ESCAPE:
            self.Dismiss()
        elif event.KeyCode == wx.WXK_DOWN or (
            event.AltDown() and event.KeyCode == ord("J")
        ):
            id = self.list_ctrl.GetSelection()
            if id != wx.NOT_FOUND:
                if id < self.list_ctrl.GetCount() - 1:
                    self.list_ctrl.SetSelection(id + 1)
        elif event.KeyCode == wx.WXK_UP or (
            event.AltDown() and event.KeyCode == ord("K")
        ):
            id = self.list_ctrl.GetSelection()
            if id != wx.NOT_FOUND:
                if id > 0:
                    self.list_ctrl.SetSelection(id - 1)
        elif event.KeyCode == wx.WXK_TAB:
            return self.on_enter(event)
        else:
            event.Skip()

    def on_text(self, event):
        event.Skip()
        s = event.GetString()
        if len(s) < self.minimum_input_length:
            return
        if self.url:
            base_url = self.url
        else:
            base_url = "/select2/fields/auto.json"
        url = base_url + "?term=%s&page=1&context=&field_id=%s" % (
            s,
            self.field_id,
        )
        http = wx.GetApp().get_http(self.combo)
        response = http.get(self, url)
        tab = schjson.loads(response.str())
        if not "err" in tab or (
            "err" in tab and (tab["err"] != "nil" or tab["err"] != None)
        ):
            self.list_ctrl.Clear()
            if len(tab["results"]) > 0:
                for pos in tab["results"]:
                    self.list_ctrl.Append(pos["text"], pos["id"])
                if s:
                    self.list_ctrl.SetSelection(0)

    def set_position(self, point):
        self.point = point

    def clear(self):
        self.edit_ctrl.ChangeValue("")
        self.list_ctrl.Clear()

    def Popup(self):
        self.Show()
        self.Move(self.point)

    def Dismiss(self):
        self.Hide()
        self.combo.SetFocus()

    def Hide(self):
        self.combo.on_popup_hidden()
        super().Hide()


class Select2Base(wx.ComboCtrl, SchBaseCtrl):
    """Base class for SELECT2 widget, server interface based on select2 javascript library: https://select2.github.io/"""

    def __init__(self, parent, **kwds):
        self.popup = None
        self.button1 = None
        self.button2 = None
        self._popup_shown = False

        SchBaseCtrl.__init__(self, parent, kwds)

        data = self.param["data"][0]["attrs"]
        if "multiple" in data:
            self.multiple = True
        else:
            self.multiple = False

        if "style" in kwds:
            kwds["style"] |= wx.TE_PROCESS_ENTER
        else:
            kwds["style"] = wx.TE_PROCESS_ENTER

        if "item_id" in self.param and self.param["item_id"] != "None":
            self.item_id = [
                int(self.param["item_id"]),
            ]
            self.item_str = [
                self.param["item_str"],
            ]
        else:
            self.item_id = [
                -1,
            ]
            self.item_str = [
                "",
            ]

        kwds["size"] = (438, -1)

        wx.ComboCtrl.__init__(self, parent, **kwds)

        if self.GetTextCtrl():
            self.GetTextCtrl().SetForegroundColour(wx.Colour(0, 0, 0))

        self.Bind(wx.EVT_CHAR, self.on_char)

        if self.item_str:
            if self.multiple:
                self.SetValue(self.item_str[0])
            else:
                self.SetValue(self.item_str[0])

        self.Bind(wx.EVT_KEY_DOWN, self.on_key_down_base)

        self.GetParent().get_parent_page().register_signal(self, "return_new_row")
        self.GetParent().get_parent_page().register_signal(self, "return_updated_row")

    def return_updated_row(self, **argv):
        self.set_value(argv["id"], argv["title"])

    def return_new_row(self, **argv):
        self.set_value(argv["id"], argv["title"])

    def init(self, button1, button2):
        self.button1 = button1
        self.button2 = button2

    def on_key_down_base(self, event):
        if event.GetKeyCode() == wx.WXK_TAB:
            if event.ShiftDown():
                self.GetParent().GetParent().Navigate(self.GetParent(), True)
            else:
                self.GetParent().GetParent().Navigate(self.GetParent(), False)
        elif event.GetKeyCode() == wx.WXK_F2:
            self.button1.on_click(event)
        elif event.GetKeyCode() == wx.WXK_INSERT:
            self.button2.on_click(event)
        else:
            event.Skip()

    def set_value(self, item_id, item_str):
        """Set value of field

        Args:
            item_id - table row id
            item_str - string representation of table row
        """
        if self.multiple:
            if not item_id in self.item_id:
                self.item_id.append(item_id)
                self.item_str.append(item_str)
            self.SetValue("; ".join(self.item_str)[2:])
        else:
            self.item_id = [
                item_id,
            ]
            self.item_str = [
                item_str,
            ]
            self.SetValue(item_str)

        def _fun():
            self.SelectNone()
            self.SetInsertionPointEnd()

        wx.CallAfter(_fun)

    def on_char(self, event):
        try:
            c = chr(event.GetUnicodeKey())
            if c in string.printable:
                if not self._popup_shown:
                    self._on_button_click()
                self.popup.edit_ctrl.AppendText(c)
        except:
            pass
        if event.KeyCode in (wx.WXK_DELETE, wx.WXK_BACK):
            self.item_id = [
                -1,
            ]
            self.item_str = [
                "",
            ]
            self.Clear()

    def _on_button_click(self):
        if not self.popup:
            pos = self.GetScreenPosition()
            pos = (pos[0], pos[1] + self.GetSize()[1])
            field_id = None
            minimum_input_length = None
            url = None
            data = self.param["data"][0]["attrs"]
            if "data-field_id" in data:
                field_id = data["data-field_id"]
            if "data-minimum-input-length" in data:
                minimum_input_length = int(data["data-minimum-input-length"])
            if "data-ajax--url" in data:
                url = data["data-ajax--url"]

            if self.GetTextCtrl():
                self.popup = Select2Popup(
                    self.GetTextCtrl(),
                    -1,
                    _("Select item"),
                    pos=pos,
                    size=(450, 400),
                    style=wx.DEFAULT_DIALOG_STYLE,
                    combo=self,
                    url=url,
                    field_id=field_id,
                    minimum_input_length=minimum_input_length,
                )
            else:
                self.popup = Select2Popup(
                    self,
                    -1,
                    _("Select item"),
                    pos=pos,
                    size=(450, 400),
                    style=wx.DEFAULT_DIALOG_STYLE,
                    combo=self,
                    url=url,
                    field_id=field_id,
                    minimum_input_length=minimum_input_length,
                )
        self.popup.clear()

        pos = self.GetScreenPosition()
        pos = (pos[0], pos[1] + self.GetSize()[1])
        pos = [pos[0], pos[1]]

        screen_dx = wx.SystemSettings.GetMetric(wx.SYS_SCREEN_X)
        screen_dy = wx.SystemSettings.GetMetric(wx.SYS_SCREEN_Y)

        popup_size = self.popup.GetSize()

        if pos[0] + popup_size[0] > screen_dx:
            pos[0] = screen_dx - popup_size[0]
        if pos[1] + popup_size[1] > screen_dy:
            pos[1] = pos[1] - self.GetSize().GetHeight() - popup_size[1]

        self.popup.set_position(pos)
        self.popup.Popup()
        self._popup_shown = True

    def on_popup_hidden(self):
        self._popup_shown = False
        self.SetFocus()

    def GetValue(self):
        """Return field value - table row id"""
        if self.multiple:
            return self.item_id
        else:
            return self.item_id[0]

    def OnButtonClick(self):
        # self.SetValue("")
        ret = self._on_button_click()
        self.popup.edit_ctrl.SetValue("")
        wx.CallAfter(self.popup.edit_ctrl.SetFocus)
        return ret

    def DoSetPopupControl(self, popup):
        pass

    def Dismiss(self):
        if self.popup:
            self.popup.Close()
            self.popup = None
        self.SetFocus()
